import re
import os
import sys
import time
import pytest
import warnings
import threading
import subprocess
import contextlib
from tempfile import mkstemp, mkdtemp, NamedTemporaryFile, _RandomNameSequence
from _multiprocessing import SemLock as _SemLock
from _multiprocessing import sem_unlink

from loky.backend import resource_tracker


_rand_name = _RandomNameSequence()


def resource_unlink(name, rtype):
    resource_tracker._CLEANUP_FUNCS[rtype](name)


def create_resource(rtype):
    if rtype == "folder":
        return mkdtemp(dir=os.getcwd())
    elif rtype == "semlock":
        name = f"test-loky-{os.getpid()}-{next(_rand_name)}"
        _SemLock(1, 1, 1, name, False)
        return name
    elif rtype == "file":
        tmpfile = NamedTemporaryFile(delete=False)
        tmpfile.close()
        return tmpfile.name
    else:
        raise ValueError(f"Resource type {rtype} not understood")


def resource_exists(name, rtype):
    if rtype in ["folder", "file"]:
        return os.path.exists(name)
    elif rtype == "semlock":
        try:
            _SemLock(1, 1, 1, name, False)
            sem_unlink(name)
            return False
        except OSError:
            return True
    else:
        raise ValueError(f"Resource type {rtype} not understood")


@contextlib.contextmanager
def captured_output(stream_name):
    """Return a context manager used by captured_stdout/stdin/stderr
    that temporarily replaces the sys stream *stream_name* with a StringIO."""
    import io

    orig_stdout = getattr(sys, stream_name)
    s = io.StringIO()

    setattr(sys, stream_name, s)
    try:
        yield getattr(sys, stream_name)
    finally:
        setattr(sys, stream_name, orig_stdout)


def captured_stderr():
    """Capture the output of sys.stderr:

    with captured_stderr() as stderr:
        print("hello", file=sys.stderr)
    self.assertEqual(stderr.getvalue(), "hello\\n")
    """
    return captured_output("stderr")


#
# Wrapper
#


class TimingWrapper:
    def __init__(self, func):
        self.func = func
        self.elapsed = None

    def __call__(self, *args, **kwds):
        t = time.time()
        try:
            return self.func(*args, **kwds)
        finally:
            self.elapsed = time.time() - t

    def assert_timing_lower_than(self, delay):
        msg = (
            f"expected duration lower than {delay:.3f}s, "
            f"got {self.elapsed:.3f}s"
        )
        assert self.elapsed < delay, msg

    def assert_timing_almost_zero(self):
        self.assert_timing_lower_than(0.1)


#
# helper functions
#


def id_sleep(x, delay=0):
    """sleep for delay seconds and return its first argument"""
    time.sleep(delay)
    return x


def check_subprocess_call(
    cmd, timeout=1, stdout_regex=None, stderr_regex=None, env=None
):
    """Runs a command in a subprocess with timeout in seconds.

    Also checks returncode is zero, stdout if stdout_regex is set, and
    stderr if stderr_regex is set.
    """
    if env is not None:
        env = {**os.environ, **env}

    proc = subprocess.Popen(
        cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, env=env, text=True
    )

    def kill_process():
        warnings.warn(f"Timeout running {cmd}")
        proc.kill()

    timer = threading.Timer(timeout, kill_process)
    try:
        timer.start()
        stdout, stderr = proc.communicate()

        if proc.returncode == -9:
            message = (
                f"Subprocess timeout after {timeout}s.\nStdout:\n{stdout}\n"
                f"Stderr:\n{stderr}"
            )
            raise TimeoutError(message)
        elif proc.returncode != 0:
            message = (
                f"Non-zero return code: {proc.returncode}.\nStdout:\n{stdout}"
                f"\nStderr:\n{stderr}"
            )
            raise ValueError(message)

        if stdout_regex is not None and not re.search(stdout_regex, stdout):
            raise ValueError(
                f"Unexpected stdout: {stdout_regex!r} does not match:"
                f"\n{stdout!r}"
            )
        if stderr_regex is not None and not re.search(stderr_regex, stderr):
            raise ValueError(
                f"Unexpected stderr: {stderr_regex!r} does not "
                f"match:\n{stderr!r}"
            )

        return stdout, stderr

    finally:
        timer.cancel()


def skip_func(msg):
    def test_func(*args, **kwargs):
        pytest.skip(msg)

    return test_func


# A decorator to run tests only when numpy is available
try:
    import numpy  # noqa F401

    def with_numpy(func):
        """A decorator to skip tests requiring numpy."""
        return func

except ImportError:

    def with_numpy(func):
        """A decorator to skip tests requiring numpy."""
        return skip_func("Test require numpy")


# A decorator to run tests only when numpy is available
try:
    from _openmp_test_helper.parallel_sum import parallel_sum  # noqa F401

    def with_parallel_sum(func):
        """A decorator to skip tests if parallel_sum is not compiled."""
        return func

    def _run_openmp_parallel_sum(*args):
        return parallel_sum(*args)

except ImportError:

    def with_parallel_sum(func):
        """A decorator to skip tests if parallel_sum is not compiled."""
        return skip_func("Test requires parallel_sum to be compiled")

    _run_openmp_parallel_sum = None


def check_python_subprocess_call(code, stdout_regex=None, timeout=10):
    cmd = [sys.executable]
    try:
        fid, filename = mkstemp(suffix="_joblib.py")
        os.close(fid)
        with open(filename, mode="w") as f:
            f.write(code)
        cmd += [filename]
        check_subprocess_call(cmd, stdout_regex=stdout_regex, timeout=timeout)
    finally:
        os.unlink(filename)


def filter_match(match):
    if sys.platform == "win32":
        return None
    return match



import signal
import multiprocessing as mp

from openmp_parallel_sum import parallel_sum


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser("Test compat openMP/multproc")
    parser.add_argument(
        "--start",
        type=str,
        default="fork",
        help="define start method tested with openMP",
    )
    args = parser.parse_args()

    parallel_sum(10)

    mp.set_start_method(args.start)
    p = mp.Process(target=parallel_sum, args=(10,))

    def raise_timeout(a, b):
        print("TIMEOUT - parallel_sum could not complete in less than a sec")
        p.terminate()

    signal.signal(signal.SIGALRM, raise_timeout)
    signal.alarm(1)
    p.start()
    p.join()
    print("done")



import inspect
from functools import partial
from cloudpickle import dumps, loads


WRAP_CACHE = {}


class CloudpickledObjectWrapper:
    def __init__(self, obj, keep_wrapper=False):
        self._obj = obj
        self._keep_wrapper = keep_wrapper

    def __reduce__(self):
        _pickled_object = dumps(self._obj)
        if not self._keep_wrapper:
            return loads, (_pickled_object,)

        return _reconstruct_wrapper, (_pickled_object, self._keep_wrapper)

    def __getattr__(self, attr):
        # Ensure that the wrapped object can be used seemlessly as the
        # previous object.
        if attr not in ["_obj", "_keep_wrapper"]:
            return getattr(self._obj, attr)
        return getattr(self, attr)


# Make sure the wrapped object conserves the callable property
class CallableObjectWrapper(CloudpickledObjectWrapper):
    def __call__(self, *args, **kwargs):
        return self._obj(*args, **kwargs)


def _wrap_non_picklable_objects(obj, keep_wrapper):
    if callable(obj):
        return CallableObjectWrapper(obj, keep_wrapper=keep_wrapper)
    return CloudpickledObjectWrapper(obj, keep_wrapper=keep_wrapper)


def _reconstruct_wrapper(_pickled_object, keep_wrapper):
    obj = loads(_pickled_object)
    return _wrap_non_picklable_objects(obj, keep_wrapper)


def _wrap_objects_when_needed(obj):
    # Function to introspect an object and decide if it should be wrapped or
    # not.
    need_wrap = "__main__" in getattr(obj, "__module__", "")
    if isinstance(obj, partial):
        return partial(
            _wrap_objects_when_needed(obj.func),
            *[_wrap_objects_when_needed(a) for a in obj.args],
            **{
                k: _wrap_objects_when_needed(v)
                for k, v in obj.keywords.items()
            },
        )
    if callable(obj):
        # Need wrap if the object is a function defined in a local scope of
        # another function.
        func_code = getattr(obj, "__code__", "")
        need_wrap |= getattr(func_code, "co_flags", 0) & inspect.CO_NESTED

        # Need wrap if the obj is a lambda expression
        func_name = getattr(obj, "__name__", "")
        need_wrap |= "<lambda>" in func_name

    if not need_wrap:
        return obj

    wrapped_obj = WRAP_CACHE.get(obj)
    if wrapped_obj is None:
        wrapped_obj = _wrap_non_picklable_objects(obj, keep_wrapper=False)
        WRAP_CACHE[obj] = wrapped_obj
    return wrapped_obj


def wrap_non_picklable_objects(obj, keep_wrapper=True):
    """Wrapper for non-picklable object to use cloudpickle to serialize them.

    Note that this wrapper tends to slow down the serialization process as it
    is done with cloudpickle which is typically slower compared to pickle. The
    proper way to solve serialization issues is to avoid defining functions and
    objects in the main scripts and to implement __reduce__ functions for
    complex classes.
    """
    # If obj is a  class, create a CloudpickledClassWrapper which instantiates
    # the object internally and wrap it directly in a CloudpickledObjectWrapper
    if inspect.isclass(obj):

        class CloudpickledClassWrapper(CloudpickledObjectWrapper):
            def __init__(self, *args, **kwargs):
                self._obj = obj(*args, **kwargs)
                self._keep_wrapper = keep_wrapper

        CloudpickledClassWrapper.__name__ = obj.__name__
        return CloudpickledClassWrapper

    # If obj is an instance of a class, just wrap it in a regular
    # CloudpickledObjectWrapper
    return _wrap_non_picklable_objects(obj, keep_wrapper=keep_wrapper)



import os
import subprocess
import sys
import gc
import ctypes
from tempfile import NamedTemporaryFile
import pytest
import warnings
import threading
from time import sleep
from multiprocessing import util, current_process, Manager
from pickle import PicklingError, UnpicklingError

import cloudpickle
from packaging.version import Version

import loky
from loky import cpu_count
from loky import get_reusable_executor
from loky.process_executor import _RemoteTraceback, TerminatedWorkerError
from loky.process_executor import BrokenProcessPool, ShutdownExecutorError
from loky.reusable_executor import _ReusablePoolExecutor
from loky.backend.context import _MAX_WINDOWS_WORKERS

try:
    import psutil
except ImportError:
    psutil = None

from ._executor_mixin import ReusableExecutorMixin
from .utils import TimingWrapper, id_sleep, check_python_subprocess_call
from .utils import filter_match

cloudpickle_version = Version(cloudpickle.__version__)

# Compat windows
if sys.platform == "win32":
    from signal import SIGTERM as SIGKILL

    libc = ctypes.cdll.msvcrt
else:
    from signal import SIGKILL
    from ctypes.util import find_library

    libc = ctypes.CDLL(find_library("c"))


try:
    import numpy as np
except ImportError:
    np = None


def clean_warning_registry():
    """Safe way to reset warnings."""
    warnings.resetwarnings()
    registry_name = "__warningregistry__"
    for mod in list(sys.modules.values()):
        warning_registry = getattr(mod, registry_name, None)
        if warning_registry is not None:
            warning_registry.clear()


def wait_dead(worker, n_tries=1000, delay=0.001):
    """Wait for process pid to die"""
    for _ in range(n_tries):
        if worker.exitcode is not None:
            return
        sleep(delay)
    raise RuntimeError(
        f"Process {worker.pid} failed to die for at least "
        f"{delay * n_tries:0.3f}s"
    )


def crash():
    """Induces a segfault"""
    import faulthandler

    faulthandler._sigsegv()


def exit():
    """Induces a sys exit with exitcode 0"""
    sys.exit(0)


def c_exit(exitcode=0):
    """Induces a libc exit with exitcode 0"""
    libc.exit(exitcode)


def sleep_then_check_pids_exist(arg):
    """Sleep for some time and the check if all the passed pids exist"""
    time, pids = arg
    sleep(time)
    res = True
    for p in pids:
        res &= psutil.pid_exists(p)
    return res


def kill_friend(pid, delay=0):
    """Function that send SIGKILL at process pid"""
    sleep(delay)
    try:
        os.kill(pid, SIGKILL)
    except (PermissionError, ProcessLookupError) as e:
        if psutil.pid_exists(pid):
            util.debug("Fail to kill an alive process?!?")
            raise e
        util.debug(f"process {pid} was already dead")


def raise_error(etype=UnpicklingError, message=None):
    """Function that raises an Exception in process"""
    raise etype(message)


def return_instance(cls):
    """Function that returns a instance of cls"""
    return cls()


class SayWhenError(ValueError):
    pass


def exception_throwing_generator(total, when):
    for i in range(total):
        if i == when:
            raise SayWhenError("Somebody said when")
        yield i


def do_nothing(arg):
    """Function that return True, test passing argument"""
    return True


class CrashAtPickle:
    """Bad object that triggers a segfault at pickling time."""

    def __reduce__(self):
        crash()


class CrashAtUnpickle:
    """Bad object that triggers a segfault at unpickling time."""

    def __reduce__(self):
        return crash, ()


class ExitAtPickle:
    """Bad object that triggers a segfault at pickling time."""

    def __reduce__(self):
        exit()


class ExitAtUnpickle:
    """Bad object that triggers a process exit at unpickling time."""

    def __reduce__(self):
        return exit, ()


class CExitAtPickle:
    """Bad object that triggers a segfault at pickling time."""

    def __reduce__(self):
        c_exit()


class CExitAtUnpickle:
    """Bad object that triggers a process exit at unpickling time."""

    def __reduce__(self):
        return c_exit, ()


class ErrorAtPickle:
    """Bad object that raises an error at pickling time."""

    def __init__(self, fail=True):
        self.fail = fail

    def __reduce__(self):
        if self.fail:
            raise PicklingError("Error in pickle")
        else:
            return id, (42,)


class ErrorAtUnpickle:
    """Bad object that triggers a process exit at unpickling time."""

    def __init__(self, etype=UnpicklingError, message="the error message"):
        self.etype = etype
        self.message = message

    def __reduce__(self):
        return raise_error, (self.etype, self.message)


class CrashAtGCInWorker:
    """Bad object that triggers a segfault at call item GC time"""

    def __del__(self):
        if current_process().name != "MainProcess":
            crash()


class CExitAtGCInWorker:
    """Exit worker at call item GC time"""

    def __del__(self):
        if current_process().name != "MainProcess":
            c_exit()


class TestExecutorDeadLock(ReusableExecutorMixin):
    crash_cases = [
        # Check problem occuring while pickling a task in
        (id, (ExitAtPickle(),), PicklingError, None),
        (id, (ErrorAtPickle(),), PicklingError, None),
        # Check problem occuring while unpickling a task on workers
        (id, (ExitAtUnpickle(),), BrokenProcessPool, "SystemExit"),
        (id, (CExitAtUnpickle(),), TerminatedWorkerError, r"EXIT\(0\)"),
        (id, (ErrorAtUnpickle(),), BrokenProcessPool, "UnpicklingError"),
        (id, (CrashAtUnpickle(),), TerminatedWorkerError, "SIGSEGV"),
        # Check problem occuring during function execution on workers
        (crash, (), TerminatedWorkerError, "SIGSEGV"),
        (exit, (), SystemExit, None),
        (c_exit, (), TerminatedWorkerError, r"EXIT\(0\)"),
        (raise_error, (RuntimeError,), RuntimeError, None),
        # Check problem occuring while pickling a task result
        # on workers
        (return_instance, (CrashAtPickle,), TerminatedWorkerError, "SIGSEGV"),
        (return_instance, (ExitAtPickle,), SystemExit, None),
        (
            return_instance,
            (CExitAtPickle,),
            TerminatedWorkerError,
            r"EXIT\(0\)",
        ),
        (return_instance, (ErrorAtPickle,), PicklingError, None),
        # Check problem occuring while unpickling a task in
        # the result_handler thread
        (return_instance, (ExitAtUnpickle,), BrokenProcessPool, "SystemExit"),
        (
            return_instance,
            (ErrorAtUnpickle,),
            BrokenProcessPool,
            "UnpicklingError",
        ),
    ]

    @pytest.mark.parametrize("func, args, expected_err, match", crash_cases)
    def test_crashes(self, func, args, expected_err, match):
        """Test various reusable_executor crash handling"""
        executor = get_reusable_executor(max_workers=2)
        res = executor.submit(func, *args)

        match_err = None
        if expected_err is TerminatedWorkerError:
            match_err = filter_match(match)
            match = None
        with pytest.raises(expected_err, match=match_err) as exc_info:
            res.result()

        # For remote traceback, ensure that the cause contains the original
        # error
        if match is not None:
            with pytest.raises(_RemoteTraceback, match=match):
                raise exc_info.value.__cause__

    @pytest.mark.parametrize("func, args, expected_err, match", crash_cases)
    def test_in_callback_submit_with_crash(
        self, func, args, expected_err, match
    ):
        """Test the recovery from callback crash"""
        executor = get_reusable_executor(max_workers=2, timeout=12)

        def in_callback_submit(future):
            future2 = get_reusable_executor(max_workers=2, timeout=12).submit(
                func, *args
            )
            # Store the future of the job submitted in the callback to make it
            # easy to introspect.
            future.callback_future = future2
            future.callback_done.set()

        # Make sure the first submitted job last a bit to make sure that
        # the callback will be called in the queue manager thread and not
        # immediately in the main thread.
        delay = 0.1
        f = executor.submit(id_sleep, 42, delay)
        f.callback_done = threading.Event()
        f.add_done_callback(in_callback_submit)
        assert f.result() == 42
        if not f.callback_done.wait(timeout=3):
            raise AssertionError("callback not done before timeout")

        match_err = None
        if expected_err is TerminatedWorkerError:
            match_err = filter_match(match)
            match = None
        with pytest.raises(expected_err, match=match_err) as exc_info:
            f.callback_future.result()

        # For remote traceback, ensure that the cause contains the original
        # error
        if match is not None:
            with pytest.raises(_RemoteTraceback, match=match):
                raise exc_info.value.__cause__

    def test_callback_crash_on_submit(self):
        """Errors in the callback execution directly in queue manager thread.

        This case can break the process executor and we want to make sure
        that we can detect the issue and recover by calling
        get_reusable_executor.
        """
        executor = get_reusable_executor(max_workers=2)

        # Make sure the first submitted job last a bit to make sure that
        # the callback will be called in the queue manager thread and not
        # immediately in the main thread.
        delay = 0.1
        f = executor.submit(id_sleep, 42, delay)
        f.add_done_callback(lambda _: exit())
        assert f.result() == 42
        assert executor.submit(id_sleep, 42, 0.1).result() == 42

        executor = get_reusable_executor(max_workers=2)
        f = executor.submit(id_sleep, 42, delay)
        f.add_done_callback(lambda _: raise_error())
        assert f.result() == 42
        assert executor.submit(id_sleep, 42, 0.0).result() == 42

    def test_deadlock_kill(self):
        """Test deadlock recovery for reusable_executor"""
        executor = get_reusable_executor(max_workers=1, timeout=None)
        # trigger the spawning of the worker process
        executor.submit(sleep, 0.1)
        worker = next(iter(executor._processes.values()))
        with pytest.warns(UserWarning) as recorded_warnings:
            executor = get_reusable_executor(max_workers=2, timeout=None)
        assert len(recorded_warnings) == 1
        expected_msg = (
            "Trying to resize an executor with running jobs:"
            " waiting for jobs completion before resizing."
        )
        assert recorded_warnings[0].message.args[0] == expected_msg
        os.kill(worker.pid, SIGKILL)
        wait_dead(worker)

        # wait for the executor to be able to detect the issue and set itself
        # in broken state:
        sleep(0.5)
        with pytest.raises(
            TerminatedWorkerError, match=filter_match("SIGKILL")
        ):
            executor.submit(id_sleep, 42, 0.1).result()

        # the get_reusable_executor factory should be able to create a new
        # working instance
        executor = get_reusable_executor(max_workers=2, timeout=None)
        assert executor.submit(id_sleep, 42, 0.0).result() == 42

    @pytest.mark.parametrize("n_proc", [1, 2, 5, 13])
    def test_crash_races(self, n_proc):
        """Test the race conditions in reusable_executor crash handling"""
        pytest.importorskip("psutil")  # required for kill_friend & co

        if sys.platform == "win32" and n_proc > 5:
            pytest.skip(
                "On win32, the paging size can be too small to import numpy "
                "multiple times in the sub-processes (imported when loading "
                "this file). Skipping while no better solution is found. See "
                "https://github.com/joblib/loky/issues/279 for more details."
            )

        # Test for external crash signal comming from neighbor
        # with various race setup
        executor = get_reusable_executor(max_workers=n_proc, timeout=None)
        executor.map(id, range(n_proc))  # trigger the creation of the workers
        pids = list(executor._processes.keys())
        assert len(pids) == n_proc
        assert None not in pids
        res = executor.map(
            sleep_then_check_pids_exist,
            [(0.0001 * (j // 2), pids) for j in range(2 * n_proc)],
        )
        assert all(res)
        with pytest.raises(
            TerminatedWorkerError, match=filter_match("SIGKILL")
        ):
            res = executor.map(kill_friend, pids[::-1])
            list(res)

    def test_imap_handle_iterable_exception(self):
        # The catch of the errors in imap generation depend on the
        # builded version of python
        executor = get_reusable_executor(max_workers=2)
        with pytest.raises(SayWhenError):
            executor.map(
                id_sleep, exception_throwing_generator(10, 3), chunksize=1
            )

        # SayWhenError seen at start of problematic chunk's results
        executor = get_reusable_executor(max_workers=2)
        with pytest.raises(SayWhenError):
            executor.map(
                id_sleep, exception_throwing_generator(20, 7), chunksize=2
            )

        executor = get_reusable_executor(max_workers=2)
        with pytest.raises(SayWhenError):
            executor.map(
                id_sleep, exception_throwing_generator(20, 7), chunksize=4
            )

    def test_queue_full_deadlock(self):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", UserWarning)
            # Ignore any UserWarning about the executor being resized
            executor = get_reusable_executor(max_workers=1)
            executor.submit(id, 42)

        fs_fail = [
            executor.submit(do_nothing, ErrorAtPickle(True))
            for _ in range(100)
        ]
        fs = [
            executor.submit(do_nothing, ErrorAtPickle(False))
            for _ in range(100)
        ]
        with pytest.raises(PicklingError):
            fs_fail[99].result()
        assert fs[99].result()

    def test_informative_error_when_fail_at_unpickle(self):
        executor = get_reusable_executor(max_workers=2)
        obj = ErrorAtUnpickle(RuntimeError, "message raised in child")
        f = executor.submit(id, obj)

        with pytest.raises(BrokenProcessPool) as exc_info:
            f.result()
        assert "RuntimeError" in str(exc_info.value.__cause__)
        assert "message raised in child" in str(exc_info.value.__cause__)

    @pytest.mark.skipif(np is None, reason="requires numpy")
    def test_numpy_dot_parent_and_child_no_freeze(self):
        """Test that no freeze happens in child process when numpy's thread
        pool is started in the parent.
        """
        a = np.random.randn(1000, 1000)
        np.dot(a, a)  # trigger the thread pool init in the parent process
        executor = get_reusable_executor(max_workers=2)
        executor.submit(np.dot, a, a).result()
        executor.shutdown(wait=True)


class TestTerminateExecutor(ReusableExecutorMixin):
    def test_shutdown_kill(self):
        """Test reusable_executor termination handling"""
        from itertools import repeat

        executor = get_reusable_executor(max_workers=5)
        res1 = executor.map(id_sleep, range(100), repeat(0.001))
        res2 = executor.map(id_sleep, range(100), repeat(1))
        assert list(res1) == list(range(100))

        shutdown = TimingWrapper(executor.shutdown)
        shutdown(wait=True, kill_workers=True)
        assert shutdown.elapsed < 5

        # We should get an error as the executor shutdowned before we fetched
        # all the results from the long running operation.
        with pytest.raises(ShutdownExecutorError):
            list(res2)

    def test_shutdown_deadlock(self):
        """Test recovery if killed after resize call"""
        pytest.importorskip("psutil")  # required for kill_friend & co

        # Test the executor.shutdown call do not cause deadlock
        executor = get_reusable_executor(max_workers=2, timeout=None)
        executor.map(id, range(2))  # start the worker processes
        executor.submit(
            kill_friend, (next(iter(executor._processes.keys())), 0.0)
        )
        sleep(0.01)
        executor.shutdown(wait=True)

    def test_kill_workers_on_new_options(self):
        # submit a long running job with no timeout
        executor = get_reusable_executor(max_workers=2, timeout=None)
        f = executor.submit(sleep, 10000)

        # change the constructor parameter while requesting not to wait
        # for the long running task to complete (the workers will get
        # shutdown forcibly)
        executor = get_reusable_executor(
            max_workers=2, timeout=5, kill_workers=True
        )
        with pytest.raises(ShutdownExecutorError):
            f.result()
        f2 = executor.submit(id_sleep, 42, 0)
        assert f2.result() == 42

    @pytest.mark.parametrize(
        "bad_object, match",
        [(CrashAtGCInWorker, "SIGSEGV"), (CExitAtGCInWorker, r"EXIT\(0\)")],
    )
    def test_call_item_gc_crash_or_exit(self, bad_object, match):
        executor = get_reusable_executor(max_workers=1)
        bad_object = bad_object()
        f = executor.submit(id, bad_object)

        # The worker will successfully send back its result to the master
        # process before crashing so this future can always be collected:
        assert f.result() is not None

        # The executor should automatically detect that the worker has crashed
        # when processing subsequently dispatched tasks:
        with pytest.raises(TerminatedWorkerError, match=filter_match(match)):
            executor.submit(gc.collect).result()
            for _ in executor.map(sleep, [0.1] * 100):
                pass

    def test_sigkill_shutdown_leaks_workers(self):
        # Create a parent process which will report its workers pids
        # and sigkill itself.
        code = """if True:
        import os

        import loky
        import psutil

        parent_pid = os.getpid()
        with loky.get_reusable_executor(timeout=1, kill_workers=True) as p:
            list(p.map(lambda x: x, range(100)))
            for pid in p._processes.keys():
                print(f'worker_pid:{pid}')
            print(f'parent_pid:{parent_pid}')
            psutil.Process(os.getpid()).kill()
        """
        with NamedTemporaryFile(
            mode="w", suffix="_joblib.py", delete=True
        ) as f:
            f.write(code)
            f.flush()
            cmd = [sys.executable, f.name]
            out = subprocess.run(
                cmd,
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                text=True,
                timeout=60,
            )
            running_workers = []
            for line in out.stdout.split("\n"):
                if line.startswith("worker_pid"):
                    worker_pid = int(line.split(":")[1])
                    try:
                        psutil.Process(worker_pid)
                        running_workers.append(worker_pid)
                    except psutil.NoSuchProcess:
                        pass
            assert not len(
                running_workers
            ), f"There are running workers left: {running_workers}"


class TestResizeExecutor(ReusableExecutorMixin):
    def test_reusable_executor_resize(self):
        """Test reusable_executor resizing"""
        # required by sleep_then_check_pids_exist
        pytest.importorskip("psutil")

        executor = get_reusable_executor(max_workers=2, timeout=None)
        executor.map(id, range(2))

        # Decreasing the executor should drop a single process and keep one of
        # the old one as it is still in a good shape. The resize should not
        # occur while there are on going works.
        pids = list(executor._processes.keys())
        res1 = executor.submit(sleep_then_check_pids_exist, (0.3, pids))
        clean_warning_registry()
        with warnings.catch_warnings(record=True) as w:
            # Cause all warnings to always be triggered.
            warnings.simplefilter("always")
            executor = get_reusable_executor(max_workers=1, timeout=None)
            assert len(w) == 1
            expected_msg = "Trying to resize an executor with running jobs"
            assert expected_msg in str(w[0].message)
            assert res1.result(), (
                "Resize should wait for current processes " " to finish"
            )
            assert len(executor._processes) == 1
            assert next(iter(executor._processes.keys())) in pids

        # Requesting the same number of process should not impact the executor
        # nor kill the processed
        old_pid = next(iter(executor._processes.keys()))
        unchanged_executor = get_reusable_executor(max_workers=1, timeout=None)
        assert len(unchanged_executor._processes) == 1
        assert unchanged_executor is executor
        assert next(iter(unchanged_executor._processes.keys())) == old_pid

        # Growing the executor again should add a single process and keep the
        # old one as it is still in a good shape
        executor = get_reusable_executor(max_workers=2, timeout=None)
        assert len(executor._processes) == 2
        assert old_pid in list(executor._processes.keys())

    @pytest.mark.parametrize("reuse", [True, False])
    @pytest.mark.parametrize("kill_workers", [True, False])
    def test_reusable_executor_resize_many_times(self, kill_workers, reuse):
        # Tentative non-regression test for a deadlock when shutting down
        # the workers of an executor prior to resizing it.
        kwargs = {
            "timeout": None,
            "kill_workers": kill_workers,
            "reuse": reuse,
        }
        with warnings.catch_warnings(record=True):
            # Cause all warnings to always be triggered.
            warnings.simplefilter("always")
            for size in [12, 2, 1, 12, 6, 1, 8, 5]:
                executor = get_reusable_executor(max_workers=size, **kwargs)
                executor.map(sleep, [0.01] * 6)
                # Do not wait for the tasks to complete.
            executor.shutdown()

    def test_kill_after_resize_call(self):
        """Test recovery if killed after resize call"""
        pytest.importorskip("psutil")  # required for kill_friend & co

        # Test the executor resizing called before a kill arrive
        executor = get_reusable_executor(max_workers=2, timeout=None)
        executor.map(id, range(2))  # trigger the creation of worker processes
        pid = next(iter(executor._processes.keys()))
        executor.submit(kill_friend, (pid, 0.1))

        with pytest.warns(UserWarning) as recorded_warnings:
            warnings.simplefilter("always")
            executor = get_reusable_executor(max_workers=1, timeout=None)
        assert len(recorded_warnings) == 1
        expected_msg = (
            "Trying to resize an executor with running jobs:"
            " waiting for jobs completion before resizing."
        )
        assert recorded_warnings[0].message.args[0] == expected_msg
        assert executor.submit(id_sleep, 42, 0.0).result() == 42
        executor.shutdown()

    def test_resize_after_timeout(self):
        with warnings.catch_warnings(record=True) as recorded_warnings:
            warnings.simplefilter("always")
            executor = get_reusable_executor(max_workers=2, timeout=0.001)
            assert executor.submit(id_sleep, 42, 0.0).result() == 42
            sleep(0.1)
            executor = get_reusable_executor(max_workers=8, timeout=0.001)
            assert executor.submit(id_sleep, 42, 0.0).result() == 42
            sleep(0.1)
            executor = get_reusable_executor(max_workers=2, timeout=0.001)
            assert executor.submit(id_sleep, 42, 0.0).result() == 42

        if len(recorded_warnings) > 1:
            expected_msg = "A worker stopped"
            assert expected_msg in recorded_warnings[0].message.args[0]


class TestGetReusableExecutor(ReusableExecutorMixin):
    def test_invalid_process_number(self):
        """Raise error on invalid process number"""

        with pytest.raises(ValueError):
            get_reusable_executor(max_workers=0)

        with pytest.raises(ValueError):
            get_reusable_executor(max_workers=-1)

        executor = get_reusable_executor()
        with pytest.raises(ValueError):
            executor._resize(max_workers=None)

    @pytest.mark.skipif(sys.platform == "win32", reason="No fork on windows")
    def test_invalid_context(self):
        """Raise error on invalid context"""

        with pytest.warns(UserWarning):
            with pytest.raises(ValueError):
                get_reusable_executor(max_workers=2, context="fork")

    def test_pass_start_method_name_as_context(self):
        executor = get_reusable_executor(max_workers=2, context="loky")
        assert executor.submit(id, 42).result() >= 0

        with pytest.raises(ValueError):
            get_reusable_executor(max_workers=2, context="bad_start_method")

    def test_interactively_defined_executor_no_main(self):
        # check that the init_main_module parameter works properly
        # when using -c option, we don't need the safeguard if __name__ ..
        # and thus test LokyProcess without the extra argument. For running
        # a script, it is necessary to use init_main_module=False.
        code = """if True:
            from loky import get_reusable_executor
            e = get_reusable_executor()

            e.submit(id, 42).result()
            print("ok")
        """
        check_python_subprocess_call(code, stdout_regex="ok")

    def test_reused_flag(self):
        executor, _ = _ReusablePoolExecutor.get_reusable_executor(
            max_workers=2
        )
        executor, reused = _ReusablePoolExecutor.get_reusable_executor(
            max_workers=2
        )
        assert reused
        executor.shutdown(kill_workers=True)
        executor, reused = _ReusablePoolExecutor.get_reusable_executor(
            max_workers=2
        )
        assert not reused

    @pytest.mark.xfail(
        cloudpickle_version >= Version("0.5.4")
        and cloudpickle_version <= Version("0.7.0"),
        reason="Known issue in cloudpickle",
    )
    # https://github.com/cloudpipe/cloudpickle/pull/240
    def test_interactively_defined_nested_functions(self):
        # Check that it's possible to call nested interactively defined
        # functions and furthermore that changing the code interactively
        # is taken into account by the single worker process.
        code = """if True:
            from loky import get_reusable_executor
            e = get_reusable_executor(max_workers=1)

            # Force a start of the children process:
            e.submit(id, 42).result()

            # Test that it's possible to call interactively defined, nested
            # functions:

            def inner_func(x):
                return -x

            def outer_func(x):
                return inner_func(x)

            assert e.submit(outer_func, 1).result() == outer_func(1) == -1

            # Test that changes to the definition of the inner function are
            # taken into account in subsequent calls to the outer function.

            def inner_func(x):
                return x

            assert e.submit(outer_func, 1).result() == outer_func(1) == 1

            print("ok")
        """
        check_python_subprocess_call(code, stdout_regex="ok")

    def test_interactively_defined_recursive_functions(self):
        # Check that it's possible to call a recursive function defined
        # in a closure.
        # Also check that calling several function that stems from the same
        # factory with different closure states results in the expected result:
        # the function definitions should not collapse in the single worker
        # process.
        code = """if True:
            from loky import get_reusable_executor
            e = get_reusable_executor(max_workers=1)

            # Force a start of the children process:
            e.submit(id, 42).result()

            def make_func(seed):
                def func(x):
                    if x <= 0:
                        return seed
                    return func(x - 1) + 1
                return func

            func = make_func(0)
            assert e.submit(func, 5).result() == func(5) == 5

            func = make_func(1)
            assert e.submit(func, 5).result() == func(5) == 6

            print("ok")
        """
        check_python_subprocess_call(code, stdout_regex="ok")

    def test_no_deadlock_on_nested_reusable_exector(self):
        # Non-regression test for a deadlock that occurred when the main
        # process was waiting for level 1 children processes to terminate while
        # they had reusable executors still managing level 2 worker processes
        # waiting for more work to arrive (untill their long worker timeout).
        # https://github.com/joblib/loky/issues/363
        code = """if True:
            from loky import get_reusable_executor
            import logging
            import multiprocessing as mp
            from time import sleep
            import faulthandler

            def enable_mp_logging():
                # Enable verbose logging for nested loky executor and workers in
                # order to make this test easier to debug in case of failure.
                log = mp.util.log_to_stderr(logging.DEBUG)
                log.handlers[0].setFormatter(logging.Formatter(
                    '[%(levelname)s:%(processName)s:%(threadName)s] %(message)s'))
                return log

            def inner_parallel_func(j):
                sleep(0.1)  # to avoid termination of parent before all workers
                mp.util.debug(f"inner_parallel_func({j}) done")

                # If this test fails, uncommenting the following line will
                # help to debug the issue:
                # faulthandler.dump_traceback_later(10, exit=True)
                return j ** 2

            def outer_parallel_func(i):
                executor = get_reusable_executor(max_workers=2)
                list(executor.map(inner_parallel_func, range(2)))
                enable_mp_logging()
                mp.util.debug(f"outer_parallel_func({i}) done")

            enable_mp_logging()

            executor = get_reusable_executor(max_workers=2)
            list(executor.map(outer_parallel_func, range(2)))
            print("ok")
        """
        check_python_subprocess_call(code, stdout_regex="ok", timeout=30)

    def test_compat_with_concurrent_futures_exception(self):
        # It should be possible to use a loky process pool executor as a dropin
        # replacement for a ProcessPoolExecutor, including when catching
        # exceptions:
        concurrent = pytest.importorskip("concurrent")
        from concurrent.futures.process import BrokenProcessPool as BPPExc

        with pytest.raises(BPPExc):
            get_reusable_executor(max_workers=2).submit(crash).result()
        e = get_reusable_executor(max_workers=2)
        f = e.submit(id, 42)

        # Ensure that loky.Future are compatible with concurrent.futures
        # (see #155)
        assert isinstance(f, concurrent.futures.Future)
        _, running = concurrent.futures.wait([f], timeout=15)
        assert len(running) == 0

    thread_configurations = [
        ("constant", "clean_start"),
        ("constant", "broken_start"),
        ("varying", "clean_start"),
        ("varying", "broken_start"),
    ]

    @pytest.mark.parametrize("workers, executor_state", thread_configurations)
    def test_reusable_executor_thread_safety(self, workers, executor_state):
        if executor_state == "clean_start":
            # Create a new shared executor and ensures that it's workers are
            # ready:
            get_reusable_executor(reuse=False).submit(id, 42).result()
        else:
            # Break the shared executor before launching the threads:
            with pytest.raises(
                TerminatedWorkerError, match=filter_match("SIGSEGV")
            ):
                executor = get_reusable_executor(reuse=False)
                executor.submit(return_instance, CrashAtPickle).result()

        def helper_func(
            output_collector, max_workers=2, n_outer_steps=5, n_inner_steps=10
        ):
            with warnings.catch_warnings():  # ignore resize warnings
                warnings.simplefilter("always")
                executor = get_reusable_executor(max_workers=max_workers)
                for _ in range(n_outer_steps):
                    results = executor.map(
                        lambda x: x**2, range(n_inner_steps)
                    )
                    expected_result = [x**2 for x in range(n_inner_steps)]
                    assert list(results) == expected_result
                output_collector.append("ok")

        if workers == "constant":
            max_workers = [2] * 10
        else:
            max_workers = [(i % 4) + 1 for i in range(10)]
        # Use the same executor with the same number of workers concurrently
        # in different threads:
        output_collector = []
        threads = [
            threading.Thread(
                target=helper_func,
                args=(output_collector, w),
                name=f"test_thread_{i:02d}_max_workers_{w:d}",
            )
            for i, w in enumerate(max_workers)
        ]

        with warnings.catch_warnings(record=True):
            for t in threads:
                t.start()
            for t in threads:
                t.join()
        assert output_collector == ["ok"] * len(threads)

    def test_reusable_executor_reuse_true(self):
        executor = get_reusable_executor(max_workers=3, timeout=42)
        executor.submit(id, 42).result()
        assert len(executor._processes) == 3
        assert executor._timeout == 42

        executor2 = get_reusable_executor(reuse=True)
        executor2.submit(id, 42).result()
        assert len(executor2._processes) == 3
        assert executor2._timeout == 42
        assert executor2 is executor

        executor3 = get_reusable_executor()
        executor3.submit(id, 42).result()
        assert len(executor3._processes) == cpu_count()
        assert executor3._timeout == 10
        assert executor3 is not executor

        executor4 = get_reusable_executor()
        assert executor4 is executor3

    def test_no_crash_on_stdin_access_in_child_subprocess_run(self):
        # non-regression test for #420.
        def call_subprocess():
            sleep(0.01)
            subprocess.run(
                [sys.executable, "-c", "import sys; sys.stdin.read(0)"]
            ).check_returncode()

        executor = get_reusable_executor(max_workers=4, timeout=2)
        executor.submit(call_subprocess).result()

    def test_faulthandler_enabled(self):
        cmd = """if 1:
            from loky import get_reusable_executor
            import faulthandler

            def f(i):
                if {expect_enabled}:
                    assert faulthandler.is_enabled()
                else:
                    assert not faulthandler.is_enabled()
                if i == 5:
                    faulthandler._sigsegv()

            if {enable_faulthandler_via_initializer}:
                executor = get_reusable_executor(max_workers=2, initializer=faulthandler.enable)
            else:
                executor = get_reusable_executor(max_workers=2)

            list(executor.map(f, range(10)))
            # This should always trigger a crash.
        """

        def check_faulthandler_output(
            expect_enabled=True, enable_faulthandler_via_initializer=False
        ):
            p = subprocess.Popen(
                [
                    sys.executable,
                    "-c",
                    cmd.format(
                        expect_enabled=expect_enabled,
                        enable_faulthandler_via_initializer=enable_faulthandler_via_initializer,
                    ),
                ],
                stderr=subprocess.PIPE,
                stdout=subprocess.PIPE,
            )
            p.wait()
            out, err = p.communicate()

            # The worker is always expected to crash, irrespective of whether
            # faulthandler is enabled or not.
            assert p.returncode == 1, out.decode()

            if expect_enabled:
                assert b"Current thread" in err, err.decode()

        original_pythonfaulthandler_env = os.environ.pop(
            "PYTHONFAULTHANDLER", None
        )
        try:
            # Case 1: faulthandler should be automatically enabled by default.
            check_faulthandler_output(expect_enabled=True)

            # Case 2: faulthandler should also be enabled when
            # PYTHONFAULTHANDLER=1 is set.
            os.environ["PYTHONFAULTHANDLER"] = "1"
            check_faulthandler_output(expect_enabled=True)

            # Case 3: faulthandler should not be enabled when
            # PYTHONFAULTHANDLER=0 is set explicitly.
            os.environ["PYTHONFAULTHANDLER"] = "0"
            check_faulthandler_output(expect_enabled=False)

            # Case 4: faulthandler can also be enabled manually via the initializer.
            del os.environ["PYTHONFAULTHANDLER"]
            check_faulthandler_output(
                expect_enabled=True, enable_faulthandler_via_initializer=True
            )
        finally:
            if original_pythonfaulthandler_env is None:
                os.environ.pop("PYTHONFAULTHANDLER", None)
            else:
                os.environ["PYTHONFAULTHANDLER"] = (
                    original_pythonfaulthandler_env
                )

    @pytest.mark.parametrize("n_cpus", [1, 2, 4])
    def test_reusable_executor_queue_size(self, n_cpus):
        # Check that the reusable executor can be lauched with more workers
        # than the number of CPUs with no limitation on the queue_size.
        #
        # Non-regression test for https://github.com/joblib/loky/issues/396.

        get_reusable_executor().shutdown()

        import os

        original_n_cpus = os.environ.get("LOKY_MAX_CPU_COUNT", None)
        try:
            os.environ["LOKY_MAX_CPU_COUNT"] = f"{n_cpus}"

            n_workers = 2 * n_cpus + 2

            manager = Manager()
            barrier = manager.Barrier(n_workers, timeout=1)

            def check_running(i, barrier):
                print(f"TASK {i} running")
                barrier.wait()
                return True

            executor = get_reusable_executor(max_workers=n_workers)
            futures = [
                executor.submit(check_running, i, barrier)
                for i in range(n_workers)
            ]
            assert all(r.result() for r in futures)
        finally:
            manager.shutdown()
            executor.shutdown()
            if original_n_cpus is None:
                os.environ.pop("LOKY_MAX_CPU_COUNT", None)
            else:
                os.environ["LOKY_MAX_CPU_COUNT"] = original_n_cpus


class TestExecutorInitializer(ReusableExecutorMixin):
    def _initializer(self, x):
        loky._initialized_state = x

    def _test_initializer(self, delay=0):
        sleep(delay)
        return getattr(loky, "_initialized_state", "uninitialized")

    def test_reusable_initializer(self):
        executor = get_reusable_executor(
            max_workers=2, initializer=self._initializer, initargs=("done",)
        )

        assert executor.submit(self._test_initializer).result() == "done"

        # when the initializer change, the executor is re-spawned
        executor = get_reusable_executor(
            max_workers=2, initializer=self._initializer, initargs=(42,)
        )

        assert executor.submit(self._test_initializer).result() == 42

        # With reuse=True, the executor use the same initializer
        executor = get_reusable_executor(max_workers=4, reuse=True)
        for x in executor.map(self._test_initializer, delay=0.1):
            assert x == 42

        # With reuse='auto', the initializer is not used anymore
        executor = get_reusable_executor(max_workers=4)
        for x in executor.map(self._test_initializer, delay=0.1):
            assert x == "uninitialized"

    # XXX: not sure if #369 has been fixed or not. Let's not make it XFAIL
    # for now to see if it is still flaky on CI.
    # @pytest.mark.xfail(reason="https://github.com/joblib/loky/issues/369")
    def test_error_in_nested_call_keeps_resource_tracker_silent(self):
        # Safety smoke test: test that nested parallel calls don't yield noisy
        # resource_tracker outputs when the grandchild errors out.
        cmd = """if 1:
            from loky import get_reusable_executor


            def raise_error(i):
                raise ValueError


            def nested_loop(f):
                executor = get_reusable_executor(max_workers=2)
                list(executor.map(f, range(10))


            executor = get_reusable_executor(max_workers=2)
            list(executor.map(nested_loop, [raise_error])
        """
        p = subprocess.Popen(
            [sys.executable, "-c", cmd],
            stderr=subprocess.PIPE,
            stdout=subprocess.PIPE,
        )
        p.wait()
        out, err = p.communicate()
        assert p.returncode == 1, out.decode()
        assert b"resource_tracker" not in err, err.decode()


def test_no_crash_max_workers_on_windows():
    # Check that loky's reusable process pool executor does not crash when the
    # user asks for more workers than the maximum number of workers supported
    # by the platform.

    # To make sure we have the proper size for the call_queue, we
    # shutdown the existing executor.
    # See https://github.com/joblib/loky#396
    get_reusable_executor().shutdown()

    # Note: on overloaded CI hosts, spawning many processes can take a long
    # time. We need to increase the timeout to avoid spurious failures when
    # making assertions on `len(executor._processes)`.
    idle_worker_timeout = 5 * 60
    with warnings.catch_warnings(record=True) as record:
        executor = get_reusable_executor(
            max_workers=_MAX_WINDOWS_WORKERS + 1,
            timeout=idle_worker_timeout,
        )
        assert executor.submit(lambda: None).result() is None
    if sys.platform == "win32":
        assert len(record) == 1
        assert "max_workers" in str(record[0].message)
        assert len(executor._processes) == _MAX_WINDOWS_WORKERS
    else:
        assert len(record) == 0
        assert len(executor._processes) == _MAX_WINDOWS_WORKERS + 1

    # Downsizing should never raise a warning.
    before_downsizing_executor = executor
    with warnings.catch_warnings(record=True) as record:
        executor = get_reusable_executor(
            max_workers=_MAX_WINDOWS_WORKERS, timeout=idle_worker_timeout
        )
        assert executor.submit(lambda: None).result() is None

    # No warning on any OS when max_workers does not exceed the limit.
    assert len(record) == 0, [w.message for w in record]
    assert before_downsizing_executor is executor
    assert len(executor._processes) == _MAX_WINDOWS_WORKERS

    executor.shutdown()



"""Tests for the ResourceTracker class"""

import errno
import gc
import os
import pytest
import re
import signal
import subprocess
import sys
import time
import warnings
import weakref

from loky import ProcessPoolExecutor
import loky.backend.resource_tracker as resource_tracker
from loky.backend.context import get_context

from .utils import resource_unlink, create_resource, resource_exists


def _resource_unlink(name, rtype):
    resource_tracker._CLEANUP_FUNCS[rtype](name)


def get_rtracker_fd():
    resource_tracker.ensure_running()
    return resource_tracker._resource_tracker._fd


class TestResourceTracker:
    @pytest.mark.parametrize("rtype", ["file", "folder", "semlock"])
    def test_resource_utils(self, rtype):
        # Check that the resouce utils work as expected in the main process
        if sys.platform == "win32" and rtype == "semlock":
            pytest.skip("no semlock on windows")
        name = create_resource(rtype)
        assert resource_exists(name, rtype)
        resource_unlink(name, rtype)
        assert not resource_exists(name, rtype)

    def test_child_retrieves_resource_tracker(self):

        # First simple fd retrieval check (see #200)
        # checking fd only work on posix for now
        if sys.platform != "win32":
            try:
                parent_rtracker_fd = get_rtracker_fd()
                executor = ProcessPoolExecutor(max_workers=2)
                child_rtracker_fd = executor.submit(get_rtracker_fd).result()

                assert child_rtracker_fd == parent_rtracker_fd
            finally:
                executor.shutdown()

        # Register a resource in the parent process, and un-register it in the
        # child process. If the two processes do not share the same
        # resource_tracker, a cache KeyError should be printed in stderr.
        cmd = """if 1:
        import os, sys

        from loky import ProcessPoolExecutor
        from loky.backend import resource_tracker
        from tempfile import NamedTemporaryFile


        tmpfile = NamedTemporaryFile(delete=False)
        tmpfile.close()
        filename = tmpfile.name
        resource_tracker.VERBOSE = True

        resource_tracker.register(filename, "file")

        def maybe_unlink(name, rtype):
            # resource_tracker.maybe_unlink is actually a bound method of the
            # ResourceTracker. We need a custom wrapper to avoid object
            # serialization.
            from loky.backend import resource_tracker
            resource_tracker.maybe_unlink(name, rtype)

        print(filename)
        e = ProcessPoolExecutor(1)
        e.submit(maybe_unlink, filename, "file").result()
        e.shutdown()
        """

        p = subprocess.run(
            [sys.executable, "-E", "-c", cmd],
            capture_output=True,
            text=True,
        )
        filename = p.stdout.strip()

        pattern = f"decremented refcount of file {filename}"
        assert pattern in p.stderr
        assert "leaked" not in p.stderr

        pattern = f"KeyError: '{filename}'"
        assert pattern not in p.stderr

    # The following four tests are inspired from cpython _test_multiprocessing
    @pytest.mark.parametrize("rtype", ["file", "folder", "semlock"])
    def test_resource_tracker(self, rtype):
        #
        # Check that killing process does not leak named resources
        #
        if (sys.platform == "win32") and rtype == "semlock":
            pytest.skip("no semlock on windows")

        cmd = f"""if 1:
            import time, os, tempfile, sys
            from loky.backend import resource_tracker
            from utils import create_resource

            for _ in range(2):
                rname = create_resource("{rtype}")
                resource_tracker.register(rname, "{rtype}")
                # give the resource_tracker time to register the new resource
                time.sleep(0.5)
                sys.stdout.write(f"{{rname}}\\n")
                sys.stdout.flush()
            time.sleep(10)
        """
        env = {**os.environ, "PYTHONPATH": os.path.dirname(__file__)}
        p = subprocess.Popen(
            [sys.executable, "-c", cmd],
            stderr=subprocess.PIPE,
            stdout=subprocess.PIPE,
            env=env,
            text=True,
        )
        name1 = p.stdout.readline().rstrip()
        name2 = p.stdout.readline().rstrip()

        # subprocess holding a reference to lock1 is still alive, so this call
        # should succeed
        _resource_unlink(name1, rtype)
        p.terminate()
        p.wait()

        # wait for the resource_tracker to cleanup the leaked resources
        time.sleep(2.0)

        with pytest.raises(OSError) as ctx:
            _resource_unlink(name2, rtype)
        # docs say it should be ENOENT, but OSX seems to give EINVAL
        assert ctx.value.errno in (errno.ENOENT, errno.EINVAL)
        err = p.stderr.read()
        p.stderr.close()
        p.stdout.close()

        expected = f"resource_tracker: There appear to be 2 leaked {rtype}"
        assert re.search(expected, err) is not None

        # resource 1 is still registered, but was destroyed externally: the
        # tracker is expected to complain.
        if sys.platform == "win32":
            errno_map = {"file": 2, "folder": 3}
            expected = (
                f"resource_tracker: {re.escape(name1)}: "
                f"(WindowsError\\(({errno_map[rtype]})|FileNotFoundError)"
            )
        else:
            expected = (
                f"resource_tracker: {re.escape(name1)}: "
                f"(OSError\\({errno.ENOENT}|FileNotFoundError)"
            )
        assert re.search(expected, err) is not None

    @pytest.mark.parametrize("rtype", ["file", "folder", "semlock"])
    def test_resource_tracker_refcounting(self, rtype):
        if sys.platform == "win32" and rtype == "semlock":
            pytest.skip("no semlock on windows")

        cmd = f"""if 1:
        import os
        import tempfile
        import time
        from loky.backend import resource_tracker
        from utils import resource_unlink, create_resource, resource_exists

        resource_tracker.VERBOSE = True

        try:
            name = create_resource("{rtype}")
            assert resource_exists(name, "{rtype}")

            from loky.backend.resource_tracker import _resource_tracker
            _resource_tracker.register(name, "{rtype}")
            _resource_tracker.register(name, "{rtype}")

            # Forget all information about the resource, but do not try to
            # remove it
            _resource_tracker.unregister(name, "{rtype}")
            time.sleep(1)
            assert resource_exists(name, "{rtype}")

            _resource_tracker.register(name, "{rtype}")
            _resource_tracker.register(name, "{rtype}")
            _resource_tracker.maybe_unlink(name, "{rtype}")
            time.sleep(1)
            assert resource_exists(name, "{rtype}")

            _resource_tracker.maybe_unlink(name, "{rtype}")
            for _ in range(100):
                if not resource_exists(name, "{rtype}"):
                    break
                time.sleep(.1)
            else:
                raise AssertionError(f"{{name}} was not unlinked in time")
        finally:
            try:
                if resource_exists(name, "{rtype}"):
                    resource_unlink(name, "{rtype}")
            except NameError:
                # "name" is not defined because create_resource has failed
                pass
        """

        env = {**os.environ, "PYTHONPATH": os.path.dirname(__file__)}
        p = subprocess.run(
            [sys.executable, "-c", cmd], capture_output=True, env=env
        )
        assert p.returncode == 0, p.stderr

    def check_resource_tracker_death(self, signum, should_die):
        # bpo-31310: if the semaphore tracker process has died, it should
        # be restarted implicitly.
        from loky.backend.resource_tracker import _resource_tracker

        pid = _resource_tracker._pid
        if pid is not None:
            os.kill(pid, signal.SIGKILL)
            os.waitpid(pid, 0)
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            _resource_tracker.ensure_running()
        pid = _resource_tracker._pid

        os.kill(pid, signum)
        time.sleep(1.0)  # give it time to die

        ctx = get_context("loky")
        with warnings.catch_warnings(record=True) as all_warn:
            warnings.simplefilter("always")

            # remove unrelated MacOS warning messages first
            warnings.filterwarnings(
                "ignore", message="semaphore are broken on OSX"
            )

            sem = ctx.Semaphore()
            sem.acquire()
            sem.release()
            wr = weakref.ref(sem)
            # ensure `sem` gets collected, which triggers communication with
            # the resource_tracker
            del sem
            gc.collect()
            assert wr() is None
            if should_die:
                assert len(all_warn) == 1
                the_warn = all_warn[0]
                assert issubclass(the_warn.category, UserWarning)
                assert "resource_tracker: process died" in str(
                    the_warn.message
                )
            else:
                assert len(all_warn) == 0, [w.message for w in all_warn]

    @pytest.mark.skipif(
        sys.platform == "win32", reason="Limited signal support on Windows"
    )
    def test_resource_tracker_sigint(self):
        # Catchable signal (ignored by resource tracker)
        self.check_resource_tracker_death(signal.SIGINT, False)

    @pytest.mark.skipif(
        sys.platform == "win32", reason="Limited signal support on Windows"
    )
    def test_resource_tracker_sigterm(self):
        # Catchable signal (ignored by resource tracker)
        self.check_resource_tracker_death(signal.SIGTERM, False)

    @pytest.mark.skipif(
        sys.platform == "win32", reason="Limited signal support on Windows"
    )
    def test_resource_tracker_sigkill(self):
        # Uncatchable signal.
        self.check_resource_tracker_death(signal.SIGKILL, True)

    def test_loky_process_inherit_multiprocessing_resource_tracker(self):
        cmd = """if 1:
        from loky import get_reusable_executor
        from multiprocessing.shared_memory import SharedMemory

        def mp_rtracker_getfd():
            from multiprocessing.resource_tracker import (
                _resource_tracker as mp_resource_tracker
            )
            return mp_resource_tracker._fd


        if __name__ == '__main__':
            executor = get_reusable_executor(max_workers=1)
            # warm up
            f = executor.submit(id, 1).result()

            # loky forces the creation of the resource tracker at process
            # creation so that loky processes can inherit its file descriptor.
            parent_fd = mp_rtracker_getfd()
            child_fd = executor.submit(mp_rtracker_getfd).result()
            assert child_fd == parent_fd

            # non-regression test for #242: unlinking in a loky process a
            # shared_memory segment tracked by multiprocessing and created its
            # parent should not generate warnings.
            shm = SharedMemory(create=True, size=10)
            f = executor.submit(shm.unlink).result()

        """
        p = subprocess.run(
            [sys.executable, "-c", cmd], capture_output=True, text=True
        )
        assert not p.stdout, p.stdout
        assert not p.stderr, p.stderr



import cython
from cython.parallel import prange
from libc.stdlib cimport malloc, free

def parallel_sum(int N):
    cdef double Ysum = 0
    cdef int i
    cdef double* X = <double *>malloc(N*cython.sizeof(double))

    for i in prange(N, nogil=True):
        Ysum += X[i]

    free(X)



from loky import process_executor
from loky.backend import get_context

from ._executor_mixin import ExecutorMixin

from ._test_process_executor import (
    AsCompletedTests,
    ExecutorShutdownTest,
    ExecutorTest,
    WaitTests,
)


class ProcessPoolSpawnMixin(ExecutorMixin):
    executor_type = process_executor.ProcessPoolExecutor
    context = get_context("spawn")


class TestsProcessPoolSpawnShutdown(
    ProcessPoolSpawnMixin, ExecutorShutdownTest
):
    def _prime_executor(self):
        pass


class TestsProcessPoolSpawnWait(ProcessPoolSpawnMixin, WaitTests):
    pass


class TestsProcessPoolSpawnAsCompleted(
    ProcessPoolSpawnMixin, AsCompletedTests
):
    pass


class TestsProcessPoolSpawnExecutor(ProcessPoolSpawnMixin, ExecutorTest):
    pass



import os
import sys
from distutils.core import setup
from Cython.Build import cythonize
from distutils.extension import Extension


if sys.platform == "darwin":
    os.environ["CC"] = "gcc-4.9"
    os.environ["CXX"] = "g++-4.9"

if sys.platform != "win32":
    extra_compile_args = ["-ffast-math", "-fopenmp"]
    extra_link_args = ["-fopenmp"]
else:
    extra_compile_args = ["/openmp"]
    extra_link_args = None

ext_modules = [
    Extension(
        "parallel_sum",
        ["parallel_sum.pyx"],
        extra_compile_args=extra_compile_args,
        extra_link_args=extra_link_args,
    )
]

setup(
    name="_openmp",
    ext_modules=cythonize(ext_modules),
)



import os
import sys
import time
import pytest
import signal
import threading

from loky.backend import get_context

from .utils import TimingWrapper

loky_context = get_context("loky")

DELTA = 0.1
TIMEOUT1 = 0.1
TIMEOUT2 = 0.3


@pytest.mark.skipif(sys.platform == "win32", reason="UNIX test")
def test_semlock_failure():
    from loky.backend.synchronize import SemLock, sem_unlink

    name = "loky-test-semlock"
    try:
        sem_unlink(name)
    except FileNotFoundError:
        pass
    try:
        sl = SemLock(0, 1, 1, name=name)
        assert sl.name == name

        with pytest.raises(FileExistsError):
            SemLock(0, 1, 1, name=name)
    finally:
        # Always clean-up the test semaphore to make this test independent of
        # previous runs (successful or not).
        sem_unlink(name)

    with pytest.raises(FileNotFoundError):
        sl._semlock._rebuild(0, 0, 0, name)


def assert_sem_value_equal(sem, value):
    try:
        assert sem.get_value() == value
    except NotImplementedError:
        pass


def assert_timing_almost_equal(t1, t2=0):
    assert abs(t1 - t2) < 2e-1


class TestLock:
    def test_lock(self):
        lock = loky_context.Lock()
        assert lock.acquire()
        assert not lock.acquire(False)
        assert lock.release() is None
        with pytest.raises(ValueError) as excinfo:
            lock.release()

        assert "released too many times" in str(excinfo.value)

    def test_rlock(self):
        lock = loky_context.RLock()
        assert lock.acquire()
        assert lock.acquire()
        assert lock.acquire()
        assert lock.release() is None
        assert lock.release() is None
        assert lock.release() is None
        with pytest.raises(AssertionError) as excinfo:
            lock.release()

        assert "not owned by thread" in str(excinfo.value)

    def test_lock_context(self):
        with loky_context.Lock():
            pass


class TestSemaphore:
    def _test_semaphore(self, sem):
        assert_sem_value_equal(sem, 2)
        assert sem.acquire()
        assert_sem_value_equal(sem, 1)
        assert sem.acquire()
        assert_sem_value_equal(sem, 0)
        assert not sem.acquire(False)
        assert_sem_value_equal(sem, 0)
        assert sem.release() is None
        assert_sem_value_equal(sem, 1)
        assert sem.release() is None
        assert_sem_value_equal(sem, 2)

    def test_semaphore(self):
        sem = loky_context.Semaphore(2)
        self._test_semaphore(sem)
        assert sem.release() is None
        assert_sem_value_equal(sem, 3)
        assert sem.release() is None
        assert_sem_value_equal(sem, 4)

    @pytest.mark.skipif(
        sys.platform == "darwin", reason="OSX have borken `get_value`"
    )
    def test_bounded_semaphore(self):
        sem = loky_context.BoundedSemaphore(2)
        self._test_semaphore(sem)
        with pytest.raises(ValueError):
            sem.release()
        assert_sem_value_equal(sem, 2)

    def test_timeout(self):

        sem = loky_context.Semaphore(0)
        acquire = TimingWrapper(sem.acquire)

        assert not acquire(False)
        assert_timing_almost_equal(acquire.elapsed)

        assert not acquire(False, None)
        assert_timing_almost_equal(acquire.elapsed)

        assert not acquire(False, TIMEOUT1)
        assert_timing_almost_equal(acquire.elapsed)

        assert not acquire(True, TIMEOUT1)
        assert_timing_almost_equal(acquire.elapsed, TIMEOUT1)

        assert not acquire(True, TIMEOUT2)
        assert_timing_almost_equal(acquire.elapsed, TIMEOUT2)


class TestCondition:
    @classmethod
    def _test_notify(cls, cond, sleeping, woken, timeout=None):
        cond.acquire()
        sleeping.release()
        cond.wait(timeout)
        woken.release()
        cond.release()

    def check_invariant(self, cond):
        # this is only supposed to succeed when there are no sleepers
        try:
            sleepers = (
                cond._sleeping_count.get_value()
                - cond._woken_count.get_value()
            )
            sleepers == 0
            cond._wait_semaphore.get_value() == 0
        except NotImplementedError:
            pass

    def test_notify(self):
        cond = loky_context.Condition()
        sleeping = loky_context.Semaphore(0)
        woken = loky_context.Semaphore(0)

        p = loky_context.Process(
            target=self._test_notify, args=(cond, sleeping, woken)
        )
        p.daemon = True
        p.start()

        p = threading.Thread(
            target=self._test_notify, args=(cond, sleeping, woken)
        )
        p.daemon = True
        p.start()

        # wait for both children to start sleeping
        sleeping.acquire()
        sleeping.acquire()

        # check no process/thread has woken up
        time.sleep(DELTA)
        assert_sem_value_equal(woken, 0)

        # wake up one process/thread
        cond.acquire()
        cond.notify()
        cond.release()

        # check one process/thread has woken up
        time.sleep(DELTA)
        assert_sem_value_equal(woken, 1)

        # wake up another
        cond.acquire()
        cond.notify()
        cond.release()

        # check other has woken up
        time.sleep(DELTA)
        assert_sem_value_equal(woken, 2)

        # check state is not mucked up
        self.check_invariant(cond)
        p.join()

    def test_notify_all(self):
        cond = loky_context.Condition()
        sleeping = loky_context.Semaphore(0)
        woken = loky_context.Semaphore(0)

        # start some threads/processes which will timeout
        for _ in range(3):
            p = loky_context.Process(
                target=self._test_notify,
                args=(cond, sleeping, woken, TIMEOUT1),
            )
            p.daemon = True
            p.start()

            t = threading.Thread(
                target=self._test_notify,
                args=(cond, sleeping, woken, TIMEOUT1),
            )
            t.daemon = True
            t.start()

        # wait for them all to sleep
        for _ in range(6):
            sleeping.acquire()

        # check they have all timed out
        for _ in range(6):
            woken.acquire()
        assert_sem_value_equal(woken, 0)

        # check state is not mucked up
        self.check_invariant(cond)

        # start some more threads/processes
        for _ in range(3):
            p = loky_context.Process(
                target=self._test_notify, args=(cond, sleeping, woken)
            )
            p.daemon = True
            p.start()

            t = threading.Thread(
                target=self._test_notify, args=(cond, sleeping, woken)
            )
            t.daemon = True
            t.start()

        # wait for them to all sleep
        for _ in range(6):
            sleeping.acquire()

        # check no process/thread has woken up
        time.sleep(DELTA)
        assert_sem_value_equal(woken, 0)

        # wake them all up
        cond.acquire()
        cond.notify_all()
        cond.release()

        # check they have all woken
        for _ in range(50):
            try:
                if woken.get_value() == 6:
                    break
            except NotImplementedError:
                break
            time.sleep(DELTA)
        assert_sem_value_equal(woken, 6)

        # check state is not mucked up
        self.check_invariant(cond)

    def test_timeout(self):
        cond = loky_context.Condition()
        wait = TimingWrapper(cond.wait)
        cond.acquire()
        res = wait(TIMEOUT1)
        cond.release()

        assert not res
        assert abs(wait.elapsed - TIMEOUT1) < 1e-1

    @classmethod
    def _test_waitfor_f(cls, cond, state):
        with cond:
            state.release()
            cond.notify()
            result = cond.wait_for(lambda: state.get_value() == 5)
            if not result or state.get_value() != 5:
                sys.exit(1)

    def test_waitfor(self):
        # based on test in test/lock_tests.py
        cond = loky_context.Condition()
        state = loky_context.Semaphore(0)
        try:
            state.get_value()
        except NotImplementedError:
            pytest.skip("`sem_get_value not implemented")

        p = loky_context.Process(
            target=self._test_waitfor_f, args=(cond, state)
        )
        p.daemon = True
        p.start()

        with cond:
            result = cond.wait_for(lambda: state.get_value() == 1)
            assert result
            assert state.get_value() == 1

        for _ in range(4):
            time.sleep(0.01)
            with cond:
                state.release()
                cond.notify()

        p.join(5)
        assert not p.is_alive()
        assert p.exitcode == 0

    @classmethod
    def _test_wait_result(cls, c, pid):
        with c:
            c.notify()
        time.sleep(1)
        if pid is not None:
            os.kill(pid, signal.SIGINT)

    def test_wait_result(self):
        if sys.platform != "win32":
            pid = os.getpid()
        else:
            pid = None

        c = loky_context.Condition()
        with c:
            assert not c.wait(0)
            assert not c.wait(0.1)

            p = loky_context.Process(
                target=self._test_wait_result, args=(c, pid)
            )
            p.start()

            assert c.wait(10)
            if pid is not None:
                with pytest.raises(KeyboardInterrupt):
                    c.wait(10)
            p.join()


class TestEvent:
    @classmethod
    def _test_event(cls, event):
        time.sleep(TIMEOUT1)
        event.set()

    def test_event(self):
        event = loky_context.Event()
        wait = TimingWrapper(event.wait)

        assert not event.is_set()

        assert not wait(0.0)
        assert_timing_almost_equal(wait.elapsed)
        assert not wait(TIMEOUT1)
        assert_timing_almost_equal(wait.elapsed, TIMEOUT1)

        event.set()

        assert event.is_set()
        assert wait()
        assert_timing_almost_equal(wait.elapsed)
        assert wait(TIMEOUT1)
        assert_timing_almost_equal(wait.elapsed)

        event.clear()

        assert not event.is_set()

        p = loky_context.Process(target=self._test_event, args=(event,))
        p.daemon = True
        p.start()
        assert wait()
        p.join()



import os
import sys
import time
import pytest
import signal
import pickle
import socket
import multiprocessing as mp
from multiprocessing.connection import wait
from tempfile import mkstemp

from loky.backend import get_context
from loky.backend.context import START_METHODS
from loky.backend.utils import kill_process_tree

from .utils import (
    TimingWrapper,
    check_subprocess_call,
    with_parallel_sum,
    _run_openmp_parallel_sum,
)

if not hasattr(socket, "socketpair"):

    def socketpair():
        s1 = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s2 = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s1.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        s1.bind((socket.gethostname(), 8080))
        s1.listen(1)
        s2.connect((socket.gethostname(), 8080))
        conn, _ = s1.accept()
        return conn, s2

    socket.socketpair = socketpair

DELTA = 0.1
ctx_loky = get_context("loky")
HAVE_SEND_HANDLE = sys.platform == "win32" or (
    hasattr(socket, "CMSG_LEN")
    and hasattr(socket, "SCM_RIGHTS")
    and hasattr(socket.socket, "sendmsg")
)
HAVE_FROM_FD = hasattr(socket, "fromfd")


class TestLokyBackend:
    # loky processes
    Process = staticmethod(ctx_loky.Process)
    current_process = staticmethod(mp.current_process)
    active_children = staticmethod(mp.active_children)

    # interprocess communication objects
    Pipe = staticmethod(ctx_loky.Pipe)
    Manager = staticmethod(ctx_loky.Manager)

    # synchronization primitives
    Lock = staticmethod(ctx_loky.Lock)
    RLock = staticmethod(ctx_loky.RLock)
    Semaphore = staticmethod(ctx_loky.Semaphore)
    BoundedSemaphore = staticmethod(ctx_loky.BoundedSemaphore)
    Condition = staticmethod(ctx_loky.Condition)
    Event = staticmethod(ctx_loky.Event)
    Queue = staticmethod(ctx_loky.Queue)
    SimpleQueue = staticmethod(ctx_loky.SimpleQueue)

    @classmethod
    def teardown_class(cls):
        """Clean up the test environment from any remaining subprocesses."""
        for child_process in cls.active_children():
            kill_process_tree(child_process)

    def test_current(self):
        current = self.current_process()
        authkey = current.authkey

        assert current.is_alive()
        assert not current.daemon
        assert isinstance(authkey, bytes)
        assert authkey
        assert current.ident == os.getpid()
        assert current.exitcode is None

    def test_daemon_argument(self):
        # By default uses the current process's daemon flag.
        proc0 = self.Process(target=self._test_process)
        assert proc0.daemon == self.current_process().daemon
        proc1 = self.Process(target=self._test_process, daemon=True)
        assert proc1.daemon
        proc2 = self.Process(target=self._test_process, daemon=False)
        assert not proc2.daemon

    @classmethod
    def _test_process(cls, q, sq, *args, **kwds):
        current = cls.current_process()
        q.put(args, timeout=1)
        sq.put(args)

        q.put(kwds, timeout=1)
        q.put(current.name, timeout=1)
        q.put(bytes(current.authkey))
        q.put(current.pid)

    @pytest.mark.parametrize("context_name", ["loky", "loky_init_main"])
    def test_process(self, context_name):
        """behavior of Process variables and functional connection objects"""
        q = self.Queue()
        sq = self.SimpleQueue()
        args = (q, sq, 1, 2)
        kwargs = {"hello": 23, "bye": 2.54}
        name = "TestLokyProcess"
        ctx = get_context(context_name)
        p = ctx.Process(
            target=self._test_process, args=args, kwargs=kwargs, name=name
        )
        p.daemon = True
        current = self.current_process()

        assert p.authkey == current.authkey
        assert not p.is_alive()
        assert p.daemon
        assert p not in self.active_children()
        assert type(self.active_children()) is list
        assert p.exitcode is None

        # Make sure we do not break security
        with pytest.raises(TypeError):
            pickle.dumps(p.authkey)

        # Make sure we detect bad pickling
        with pytest.raises(RuntimeError):
            pickle.dumps(q)

        p.start()

        assert p.exitcode is None
        assert p.is_alive()
        assert p in self.active_children()

        assert q.get() == args[2:]
        assert sq.get() == args[2:]

        assert q.get() == kwargs
        assert q.get() == p.name
        assert q.get() == current.authkey
        assert q.get() == p.pid

        p.join()

        assert p.exitcode == 0
        assert not p.is_alive()
        assert p not in self.active_children()

    @classmethod
    def _test_connection(cls, conn):
        """Make sure a connection object is functional"""
        if hasattr(conn, "get"):
            conn = conn.get()
        if hasattr(conn, "accept"):
            msg = conn.recv(2)
            conn.send(msg)
        else:
            msg = conn.recv_bytes()
            conn.send_bytes(msg)
        conn.close()

    def test_socket(self):
        """sockets can be pickled at spawn and are able to send/recv"""
        server, client = socket.socketpair()

        p = self.Process(target=self._test_connection, args=(server,))
        p.start()

        client.settimeout(5)

        msg = b"42"
        client.send(msg)
        assert client.recv(2) == msg

        p.join()
        assert p.exitcode == 0

        client.close()
        server.close()

    @pytest.mark.skipif(
        not HAVE_SEND_HANDLE or not HAVE_FROM_FD,
        reason="This system cannot send handle between process"
        ". Connections object should be shared at spawning.",
    )
    def test_socket_queue(self):
        """sockets can be pickled in a queue and are able to send/recv"""
        q = self.SimpleQueue()

        p = self.Process(target=self._test_connection, args=(q,))
        p.start()

        server, client = socket.socketpair()
        q.put(server)

        msg = b"42"
        client.settimeout(5)
        client.send(msg)
        assert client.recv(2) == msg

        p.join()
        assert p.exitcode == 0

        client.close()
        server.close()

    def test_connection(self):
        """connections can be pickled at spawn and are able to send/recv"""
        parent_connection, child_connection = self.Pipe(duplex=True)

        p = self.Process(
            target=self._test_connection, args=(child_connection,)
        )
        p.start()

        msg = b"42"
        parent_connection.send(msg)
        assert parent_connection.recv() == msg

        p.join()
        assert p.exitcode == 0
        parent_connection.close()
        child_connection.close()

    @pytest.mark.skipif(
        not HAVE_SEND_HANDLE,
        reason="This system cannot send handle between. "
        "Connections object should be shared at spawning.",
    )
    def test_connection_queue(self):
        """connections can be pickled in a queue and are able to send/recv"""
        q = self.SimpleQueue()
        p = self.Process(target=self._test_connection, args=(q,))
        p.start()

        parent_connection, child_connection = self.Pipe(duplex=True)
        q.put(child_connection)

        msg = b"42"
        parent_connection.send(msg)
        assert parent_connection.recv() == msg

        p.join()
        assert p.exitcode == 0
        parent_connection.close()
        child_connection.close()

    @staticmethod
    def _test_child_env(key, queue):
        queue.put(os.environ.get(key, "not set"))

    def test_child_env_process(self):
        key = "loky_child_env_process"
        value = "loky works"
        out_queue = self.SimpleQueue()
        try:
            # Test that the environment variable is correctly copied in the
            # child process.
            os.environ[key] = value
            p = self.Process(
                target=self._test_child_env, args=(key, out_queue)
            )
            p.start()
            child_var = out_queue.get()
            p.join()

            assert child_var == value

            # Test that the environment variable is correctly overwritted by
            # using the `env` argument in Process.
            new_value = "loky rocks"
            p = self.Process(
                target=self._test_child_env,
                args=(key, out_queue),
                env={key: new_value},
            )
            p.start()
            child_var = out_queue.get()
            p.join()

            assert child_var == new_value, p.env
        finally:
            del os.environ[key]

    @classmethod
    def _test_terminate(cls, event):
        # Notify the main process that child process started
        event.set()
        time.sleep(100)

    def test_terminate(self):
        manager = self.Manager()
        event = manager.Event()

        p = self.Process(target=self._test_terminate, args=(event,))
        p.daemon = True
        p.start()

        assert p.is_alive()
        assert p in self.active_children()
        assert p.exitcode is None

        join = TimingWrapper(p.join)

        assert join(0) is None
        join.assert_timing_almost_zero()
        assert p.is_alive()

        assert join(-1) is None
        join.assert_timing_almost_zero()
        assert p.is_alive()

        # wait for child process to be fully setup
        event.wait(5)

        p.terminate()

        MAX_JOIN_TIME = 10
        if hasattr(signal, "alarm"):
            # On the Gentoo buildbot waitpid() often seems to block forever.
            # We use alarm() to interrupt it if it blocks for too long.
            def handler(*args):
                raise RuntimeError(f"join took too long: {p}")

            old_handler = signal.signal(signal.SIGALRM, handler)
            try:
                signal.alarm(MAX_JOIN_TIME)
                assert join() is None
            finally:
                signal.alarm(0)
                signal.signal(signal.SIGALRM, old_handler)
        else:
            assert join() is None

        join.assert_timing_lower_than(MAX_JOIN_TIME)

        assert not p.is_alive()
        assert p not in self.active_children()

        p.join()
        manager.shutdown()

        # XXX sometimes get p.exitcode == 0 on Windows ...
        # assert p.exitcode == -signal.SIGTERM

    def test_active_children(self):
        assert type(self.active_children()) == list

        p = self.Process(target=time.sleep, args=(DELTA,))
        assert p not in self.active_children()

        p.daemon = True
        p.start()
        assert p in self.active_children()

        p.join()
        assert p not in self.active_children()
        assert p.exitcode == 0

    @classmethod
    def _test_recursion(cls, wconn, l):
        wconn.send(l)
        if len(l) < 2:
            for i in range(2):
                p = cls.Process(
                    target=cls._test_recursion, args=(wconn, l + [i])
                )
                p.start()
                p.join()
                assert p.exitcode == 0

    def test_recursion(self):
        rconn, wconn = self.Pipe(duplex=False)
        self._test_recursion(wconn, [])

        time.sleep(DELTA)
        result = []
        while rconn.poll():
            a = rconn.recv()
            result.append(a)

        expected = [[], [0], [0, 0], [0, 1], [1], [1, 0], [1, 1]]
        assert result == expected

    @classmethod
    def _test_sentinel(cls, event):
        event.wait(30.0)

    def test_sentinel(self):
        event = self.Event()
        p = self.Process(target=self._test_sentinel, args=(event,))
        with pytest.raises(ValueError):
            p.sentinel
        p.start()
        assert not wait_for_handle(p.sentinel, timeout=0.0)
        event.set()
        p.join()
        assert p.exitcode == 0
        assert wait_for_handle(p.sentinel, timeout=1)

    @classmethod
    def _test_wait_sentinel(cls):
        from signal import SIGTERM

        time.sleep(0.1)
        os.kill(os.getpid(), SIGTERM)

    def test_wait_sentinel(self):
        p = self.Process(target=self._test_wait_sentinel)
        with pytest.raises(ValueError):
            p.sentinel
        p.start()
        assert isinstance(p.sentinel, int)
        assert not wait([p.sentinel], timeout=0.0)
        assert wait([p.sentinel], timeout=5), p.exitcode
        expected_code = 15 if sys.platform == "win32" else -15
        p.join()  # force refresh of p.exitcode
        assert p.exitcode == expected_code

    @classmethod
    def _high_number_pipe(cls):
        """Create a Pipe with 2 high numbered file descriptors"""
        fds = [fd for _ in range(50) for fd in os.pipe()]
        r, w = cls.Pipe(duplex=False)
        for fd in fds:
            os.close(fd)
        return r, w

    @classmethod
    def _test_sync_object_handling(cls, started, stop, conn, w):
        """Check validity of parents args and Create semaphores to clean up

        started, stop: Event
            make sure the main Process use lsof when this Process is setup
        conn: Connection
            an open pipe that should be closed at exit
        w: int
            fileno of the writable end of the Pipe, it should be closed
        """
        to_clean_up = [
            cls.Semaphore(0),
            cls.BoundedSemaphore(1),  # noqa: F841
            cls.Lock(),
            cls.RLock(),
            cls.Condition(),
            cls.Event(),
        ]
        started.set()
        assert conn.recv_bytes() == b"foo"
        with pytest.raises(OSError):
            os.fstat(w)
        stop.wait(5)

    def _check_fds(self, pid, w):
        """List all the open files and check no extra files are presents.

        Return a list of open named semaphores
        """
        import subprocess

        try:
            out = subprocess.check_output(
                f"lsof -a -Fftn -p {pid} -d ^txt,^cwd,^rtd".split(), text=True
            )
            lines = out.splitlines()[1:]
        except (FileNotFoundError, OSError):
            print(
                "lsof does not exist on this platform. Skip open files"
                "check."
            )
            return []

        n_pipe = 0
        named_sem = []
        unexpected = []
        for fd, t, name in zip(lines[::3], lines[1::3], lines[2::3]):
            # Check if fd is a standard IO file.
            is_std = fd in ["f0", "f1", "f2"]

            # Check if fd is a pipe.
            is_pipe = t in ["tPIPE", "tFIFO"]
            n_pipe += is_pipe

            # Check if fd is open for the rng. This can happen on different
            # platform and depending of the python version.
            is_rng = name == "n/dev/urandom"

            # Not sure if this really expected to be open or not...
            is_dev_null = name == "n/dev/null"

            # Check if fd is a semaphore or an open library. Store all the
            # named semaphore.
            is_mem = fd in ["fmem", "fDEL"]
            if sys.platform == "darwin":
                is_mem |= "n/loky-" in name
            if is_mem and "n/dev/shm/sem." in name:
                named_sem += [name[1:]]

            # No other files should be opened at this stage in the process.
            if not (is_pipe or is_std or is_rng or is_mem or is_dev_null):
                unexpected.append((fd, t, name))

        assert len(unexpected) == 0

        # there should be:
        # - one pipe for communication with main process
        # - loky's resource_tracker pipe
        # - the Connection pipe
        # - multiprocessing's resource_tracker pipe (POSIX only)
        if os.name == "posix":
            n_expected_pipes = 4
        else:
            n_expected_pipes = 3
        msg = (
            "Some pipes were not properly closed during the child process "
            "setup."
        )
        assert n_pipe == n_expected_pipes, msg

        # assert that the writable part of the Pipe (not passed to child),
        # have been properly closed.
        assert len(set(f"f{w}").intersection(lines)) == 0

        return named_sem

    def test_sync_object_handling(self):
        """Check the correct handling of semaphores and pipes with loky

        We use a Pipe object to check the stated of file descriptors in parent
        and child. To make sure there is no interference in the fd numbers, we
        use high number fd, so newly created fd should be inferior.

        To ensure we have the right number of fd in the child Process, we used
        `lsof` as it is compatible with Unix systems.
        """

        # TODO generate high numbered mp.Pipe directly
        # -> can be used on windows
        r, w = self._high_number_pipe()

        tmp_fname = "/tmp/foobar" if sys.platform != "win32" else ".foobar"
        with open(tmp_fname, "w"):
            # Process creating semaphore and pipes before stopping
            started, stop = self.Event(), self.Event()
            p = self.Process(
                target=self._test_sync_object_handling,
                args=(started, stop, r, w.fileno()),
            )
            named_sem = []
            try:
                p.start()
                assert started.wait(5), "The process took too long to start"
                r.close()
                w.send_bytes(b"foo")
                if sys.platform != "win32":
                    named_sem = self._check_fds(p.pid, w.fileno())

            finally:
                stop.set()
                p.join()

                # ensure that Pipe->r was closed when the child process exited
                with pytest.raises(IOError):
                    w.send_bytes(b"foo")
                w.close()

                if sys.platform == "linux":
                    # On linux, check that the named semaphores created in the
                    # child process have been unlinked when it terminated.
                    pid = str(os.getpid())
                    for sem in named_sem:
                        if pid not in sem:
                            assert not os.path.exists(sem), (
                                "Some named semaphore are not properly cleaned"
                                " up"
                            )

                assert p.exitcode == 0

    @with_parallel_sum
    def test_compatibility_openmp(self):
        # Use openMP before launching subprocesses. With fork backend, some fds
        # are nto correctly clean up, causing a freeze. No freeze should be
        # detected with loky.
        _run_openmp_parallel_sum(10)
        p = self.Process(target=_run_openmp_parallel_sum, args=(100,))
        p.start()
        p.join()
        assert p.exitcode == 0

    @pytest.mark.parametrize("run_file", [True, False])
    def test_interactively_define_process_no_main(self, run_file):
        # check that the init_main_module parameter works properly
        # when using -c option, we don't need the safeguard if __name__ ..
        # and thus test LokyProcess without the extra argument. For running
        # a script, it is necessary to use init_main_module=False.
        code = "\n".join(
            [
                "from loky.backend.process import LokyProcess",
                "p = LokyProcess(target=id, args=(1,), ",
                f"                init_main_module={not run_file})",
                "p.start()",
                "p.join()",
                'msg = "LokyProcess failed to load without safeguard"',
                "assert p.exitcode == 0, msg",
                'print("ok")',
            ]
        )
        cmd = [sys.executable]
        try:
            if run_file:
                fid, filename = mkstemp(suffix="_joblib.py")
                os.close(fid)
                with open(filename, mode="w") as f:
                    f.write(code)
                cmd += [filename]
            else:
                cmd += ["-c", code]
            check_subprocess_call(cmd, stdout_regex="ok", timeout=10)
        finally:
            if run_file:
                os.unlink(filename)

    def test_interactively_define_process_fail_main(self):
        # check that the default behavior of the LokyProcess is correct
        code = "\n".join(
            [
                "from loky.backend.process import LokyProcess",
                "p = LokyProcess(target=id, args=(1,),",
                "                init_main_module=True)",
                "p.start()",
                "p.join()",
                'msg = "LokyProcess succeed without safeguards"',
                "assert p.exitcode != 0, msg",
            ]
        )
        fid, filename = mkstemp(suffix="_joblib.py")
        os.close(fid)
        try:
            with open(filename, mode="w") as f:
                f.write(code)
            stdout, stderr = check_subprocess_call(
                [sys.executable, filename], timeout=10
            )
            if sys.platform == "win32":
                assert "RuntimeError:" in stderr
            else:
                assert "RuntimeError:" in stdout
        finally:
            os.unlink(filename)

    def test_loky_get_context(self):
        # check the behavior of get_context
        ctx_default = get_context()
        assert ctx_default.get_start_method() == "loky"

        ctx_loky = get_context("loky")
        assert ctx_loky.get_start_method() == "loky"

        ctx_loky_init_main = get_context("loky_init_main")
        assert ctx_loky_init_main.get_start_method() == "loky_init_main"

        with pytest.raises(ValueError):
            get_context("not_available")

    def test_interactive_contex_no_main(self):
        # Ensure that loky context is working properly
        code = "\n".join(
            [
                "from loky.backend import get_context",
                "ctx = get_context()",
                'assert ctx.get_start_method() == "loky"',
                "p = ctx.Process(target=id, args=(1,))",
                "p.start()",
                "p.join()",
                'msg = "loky context failed to load without safeguard"',
                "assert p.exitcode == 0, msg",
                'print("ok")',
            ]
        )
        try:
            fid, filename = mkstemp(suffix="_joblib.py")
            os.close(fid)
            with open(filename, mode="w") as f:
                f.write(code)
            check_subprocess_call(
                [sys.executable, filename], stdout_regex="ok", timeout=10
            )
        finally:
            os.unlink(filename)


def wait_for_handle(handle, timeout):
    if timeout is not None and timeout < 0.0:
        timeout = None
    return wait([handle], timeout)


def _run_nested_delayed(depth, delay, event):
    if depth > 0:
        p = ctx_loky.Process(
            target=_run_nested_delayed, args=(depth - 1, delay, event)
        )
        p.start()
        p.join()
    else:
        event.set()

    time.sleep(delay)


@pytest.mark.parametrize("use_psutil", [True, False])
def test_kill_process_tree(use_psutil):
    psutil = pytest.importorskip("psutil")
    event = ctx_loky.Event()
    p = ctx_loky.Process(target=_run_nested_delayed, args=(4, 1000, event))
    p.start()

    # Wait for all the processes to be launched
    if not event.wait(30):
        kill_process_tree(p, use_psutil=use_psutil)
        raise RuntimeError(
            "test_kill_process_tree was not able to launch "
            "all nested processes."
        )

    children = psutil.Process(pid=p.pid).children(recursive=True)
    kill_process_tree(p, use_psutil=use_psutil)

    # The process can take some time finishing so we should wait up to 5s
    gone, alive = psutil.wait_procs(children, timeout=5)
    msg = f"Should be no descendant left but found:\n{alive}"
    assert len(alive) == 0, msg


def _test_default_subcontext(queue):
    queue.put(mp.get_start_method())


@pytest.mark.parametrize("method", START_METHODS)
def test_default_subcontext(method):
    code = f"""if True:
        import sys

        from loky.backend.context import get_context, set_start_method
        from tests.test_loky_backend import _test_default_subcontext

        set_start_method('{method}')
        ctx = get_context()
        assert ctx.get_start_method() == '{method}'

        queue = ctx.SimpleQueue()
        p = ctx.Process(target=_test_default_subcontext, args=(queue,))
        p.start()
        p.join()
        start_method = queue.get()
        assert start_method == '{method}', start_method

        try:
            set_start_method('loky')
        except RuntimeError:
            pass
        else:
            raise AssertionError("Did not raise RuntimeError when resetting"
                                 "start_method without force")

        set_start_method(None, force=True)
        ctx = get_context()
        assert ctx.get_start_method() == 'loky'
    """

    cmd = [sys.executable, "-c", code]
    check_subprocess_call(cmd, timeout=10)

    ctx_default = get_context()
    assert ctx_default.get_start_method() == "loky"



###############################################################################
# Server process to keep track of unlinked resources, like folders and
# semaphores and clean them.
#
# author: Thomas Moreau
#
# Adapted from multiprocessing/resource_tracker.py
#  * add some VERBOSE logging,
#  * add support to track folders,
#  * add Windows support,
#  * refcounting scheme to avoid unlinking resources still in use.
#
# On Unix we run a server process which keeps track of unlinked
# resources. The server ignores SIGINT and SIGTERM and reads from a
# pipe. The resource_tracker implements a reference counting scheme: each time
# a Python process anticipates the shared usage of a resource by another
# process, it signals the resource_tracker of this shared usage, and in return,
# the resource_tracker increments the resource's reference count by 1.
# Similarly, when access to a resource is closed by a Python process, the
# process notifies the resource_tracker by asking it to decrement the
# resource's reference count by 1.  When the reference count drops to 0, the
# resource_tracker attempts to clean up the underlying resource.

# Finally, every other process connected to the resource tracker has a copy of
# the writable end of the pipe used to communicate with it, so the resource
# tracker gets EOF when all other processes have exited. Then the
# resource_tracker process unlinks any remaining leaked resources (with
# reference count above 0)

# For semaphores, this is important because the system only supports a limited
# number of named semaphores, and they will not be automatically removed till
# the next reboot.  Without this resource tracker process, "killall python"
# would probably leave unlinked semaphores.

# Note that this behavior differs from CPython's resource_tracker, which only
# implements list of shared resources, and not a proper refcounting scheme.
# Also, CPython's resource tracker will only attempt to cleanup those shared
# resources once all processes connected to the resource tracker have exited.


import os
import shutil
import sys
import signal
import warnings
from _multiprocessing import sem_unlink
from multiprocessing import util
from multiprocessing.resource_tracker import (
    ResourceTracker as _ResourceTracker,
)

from . import spawn

if sys.platform == "win32":
    import _winapi
    import msvcrt
    from multiprocessing.reduction import duplicate


__all__ = ["ensure_running", "register", "unregister"]

_HAVE_SIGMASK = hasattr(signal, "pthread_sigmask")
_IGNORED_SIGNALS = (signal.SIGINT, signal.SIGTERM)

_CLEANUP_FUNCS = {"folder": shutil.rmtree, "file": os.unlink}

if os.name == "posix":
    _CLEANUP_FUNCS["semlock"] = sem_unlink


VERBOSE = False


class ResourceTracker(_ResourceTracker):
    """Resource tracker with refcounting scheme.

    This class is an extension of the multiprocessing ResourceTracker class
    which implements a reference counting scheme to avoid unlinking shared
    resources still in use in other processes.

    This feature is notably used by `joblib.Parallel` to share temporary
    folders and memory mapped files between the main process and the worker
    processes.

    The actual implementation of the refcounting scheme is in the main
    function, which is run in a dedicated process.
    """

    def maybe_unlink(self, name, rtype):
        """Decrement the refcount of a resource, and delete it if it hits 0"""
        self.ensure_running()
        self._send("MAYBE_UNLINK", name, rtype)

    def ensure_running(self):
        """Make sure that resource tracker process is running.

        This can be run from any process.  Usually a child process will use
        the resource created by its parent."""
        with self._lock:
            if self._fd is not None:
                # resource tracker was launched before, is it still running?
                if self._check_alive():
                    # => still alive
                    return
                # => dead, launch it again
                os.close(self._fd)
                if os.name == "posix":
                    try:
                        # At this point, the resource_tracker process has been
                        # killed or crashed. Let's remove the process entry
                        # from the process table to avoid zombie processes.
                        os.waitpid(self._pid, 0)
                    except OSError:
                        # The process was terminated or is a child from an
                        # ancestor of the current process.
                        pass
                self._fd = None
                self._pid = None

                warnings.warn(
                    "resource_tracker: process died unexpectedly, "
                    "relaunching.  Some folders/sempahores might "
                    "leak."
                )

            fds_to_pass = []
            try:
                fds_to_pass.append(sys.stderr.fileno())
            except Exception:
                pass

            r, w = os.pipe()
            if sys.platform == "win32":
                _r = duplicate(msvcrt.get_osfhandle(r), inheritable=True)
                os.close(r)
                r = _r

            cmd = f"from {main.__module__} import main; main({r}, {VERBOSE})"
            try:
                fds_to_pass.append(r)
                # process will out live us, so no need to wait on pid
                exe = spawn.get_executable()
                args = [exe, *util._args_from_interpreter_flags(), "-c", cmd]
                util.debug(f"launching resource tracker: {args}")
                # bpo-33613: Register a signal mask that will block the
                # signals.  This signal mask will be inherited by the child
                # that is going to be spawned and will protect the child from a
                # race condition that can make the child die before it
                # registers signal handlers for SIGINT and SIGTERM. The mask is
                # unregistered after spawning the child.
                try:
                    if _HAVE_SIGMASK:
                        signal.pthread_sigmask(
                            signal.SIG_BLOCK, _IGNORED_SIGNALS
                        )
                    pid = spawnv_passfds(exe, args, fds_to_pass)
                finally:
                    if _HAVE_SIGMASK:
                        signal.pthread_sigmask(
                            signal.SIG_UNBLOCK, _IGNORED_SIGNALS
                        )
            except BaseException:
                os.close(w)
                raise
            else:
                self._fd = w
                self._pid = pid
            finally:
                if sys.platform == "win32":
                    _winapi.CloseHandle(r)
                else:
                    os.close(r)

    def __del__(self):
        # ignore error due to trying to clean up child process which has already been
        # shutdown on windows See https://github.com/joblib/loky/pull/450
        # This is only required if __del__ is defined
        if not hasattr(_ResourceTracker, "__del__"):
            return
        try:
            super().__del__()
        except ChildProcessError:
            pass


_resource_tracker = ResourceTracker()
ensure_running = _resource_tracker.ensure_running
register = _resource_tracker.register
maybe_unlink = _resource_tracker.maybe_unlink
unregister = _resource_tracker.unregister
getfd = _resource_tracker.getfd


def main(fd, verbose=0):
    """Run resource tracker."""
    # protect the process from ^C and "killall python" etc
    if verbose:
        util.log_to_stderr(level=util.DEBUG)

    signal.signal(signal.SIGINT, signal.SIG_IGN)
    signal.signal(signal.SIGTERM, signal.SIG_IGN)

    if _HAVE_SIGMASK:
        signal.pthread_sigmask(signal.SIG_UNBLOCK, _IGNORED_SIGNALS)

    for f in (sys.stdin, sys.stdout):
        try:
            f.close()
        except Exception:
            pass

    if verbose:
        util.debug("Main resource tracker is running")

    registry = {rtype: {} for rtype in _CLEANUP_FUNCS.keys()}
    try:
        # keep track of registered/unregistered resources
        if sys.platform == "win32":
            fd = msvcrt.open_osfhandle(fd, os.O_RDONLY)
        with open(fd, "rb") as f:
            while True:
                line = f.readline()
                if line == b"":  # EOF
                    break
                try:
                    splitted = line.strip().decode("ascii").split(":")
                    # name can potentially contain separator symbols (for
                    # instance folders on Windows)
                    cmd, name, rtype = (
                        splitted[0],
                        ":".join(splitted[1:-1]),
                        splitted[-1],
                    )

                    if cmd == "PROBE":
                        continue

                    if rtype not in _CLEANUP_FUNCS:
                        raise ValueError(
                            f"Cannot register {name} for automatic cleanup: "
                            f"unknown resource type ({rtype}). Resource type "
                            "should be one of the following: "
                            f"{list(_CLEANUP_FUNCS.keys())}"
                        )

                    if cmd == "REGISTER":
                        if name not in registry[rtype]:
                            registry[rtype][name] = 1
                        else:
                            registry[rtype][name] += 1

                        if verbose:
                            util.debug(
                                "[ResourceTracker] incremented refcount of "
                                f"{rtype} {name} "
                                f"(current {registry[rtype][name]})"
                            )
                    elif cmd == "UNREGISTER":
                        del registry[rtype][name]
                        if verbose:
                            util.debug(
                                f"[ResourceTracker] unregister {name} {rtype}: "
                                f"registry({len(registry)})"
                            )
                    elif cmd == "MAYBE_UNLINK":
                        registry[rtype][name] -= 1
                        if verbose:
                            util.debug(
                                "[ResourceTracker] decremented refcount of "
                                f"{rtype} {name} "
                                f"(current {registry[rtype][name]})"
                            )

                        if registry[rtype][name] == 0:
                            del registry[rtype][name]
                            try:
                                if verbose:
                                    util.debug(
                                        f"[ResourceTracker] unlink {name}"
                                    )
                                _CLEANUP_FUNCS[rtype](name)
                            except Exception as e:
                                warnings.warn(
                                    f"resource_tracker: {name}: {e!r}"
                                )

                    else:
                        raise RuntimeError(f"unrecognized command {cmd!r}")
                except BaseException:
                    try:
                        sys.excepthook(*sys.exc_info())
                    except BaseException:
                        pass
    finally:
        # all processes have terminated; cleanup any remaining resources
        def _unlink_resources(rtype_registry, rtype):
            if rtype_registry:
                try:
                    warnings.warn(
                        "resource_tracker: There appear to be "
                        f"{len(rtype_registry)} leaked {rtype} objects to "
                        "clean up at shutdown"
                    )
                except Exception:
                    pass
            for name in rtype_registry:
                # For some reason the process which created and registered this
                # resource has failed to unregister it. Presumably it has
                # died.  We therefore clean it up.
                try:
                    _CLEANUP_FUNCS[rtype](name)
                    if verbose:
                        util.debug(f"[ResourceTracker] unlink {name}")
                except Exception as e:
                    warnings.warn(f"resource_tracker: {name}: {e!r}")

        for rtype, rtype_registry in registry.items():
            if rtype == "folder":
                continue
            else:
                _unlink_resources(rtype_registry, rtype)

        # The default cleanup routine for folders deletes everything inside
        # those folders recursively, which can include other resources tracked
        # by the resource tracker). To limit the risk of the resource tracker
        # attempting to delete twice a resource (once as part of a tracked
        # folder, and once as a resource), we delete the folders after all
        # other resource types.
        if "folder" in registry:
            _unlink_resources(registry["folder"], "folder")

    if verbose:
        util.debug("resource tracker shut down")


def spawnv_passfds(path, args, passfds):
    if sys.platform != "win32":
        args = [arg.encode("utf-8") for arg in args]
        path = path.encode("utf-8")
        return util.spawnv_passfds(path, args, passfds)
    else:
        passfds = sorted(passfds)
        cmd = " ".join(f'"{x}"' for x in args)
        try:
            _, ht, pid, _ = _winapi.CreateProcess(
                path, cmd, None, None, True, 0, None, None, None
            )
            _winapi.CloseHandle(ht)
        except BaseException:
            pass
        return pid



# Make it possible to enable test coverage reporting for Python
# code run in children processes.
# http://coverage.readthedocs.io/en/latest/subprocess.html

import os.path as op
from sysconfig import get_path

FILE_CONTENT = """\
import coverage; coverage.process_startup()
"""

# get_path("platlib") returns the location of the `site-packages` folder:
# https://docs.python.org/3/library/sysconfig.html
filename = op.join(get_path("platlib"), "coverage_subprocess.pth")
with open(filename, mode="w") as f:
    f.write(FILE_CONTENT)

print(f"Installed subprocess coverage support: {filename}")



import sys
import time
import math
import pytest
import threading
from time import sleep
import random

from loky import TimeoutError, get_reusable_executor
from loky.backend import get_context


try:
    import psutil

    psutil_exceptions = (psutil.NoSuchProcess, psutil.AccessDenied)
except ImportError:
    psutil = None
    psutil_exceptions = ()


# Set a large timeout as it should only be reached in case of deadlocks
TIMEOUT = 40

_test_event = None


def initializer_event(event):
    """Initializer that set a global test event for test synchronization"""
    global _test_event
    _test_event = event

    # Inject some randomness in the initialization to reveal race conditions.
    if random.random() < 0.2:
        sleep(random.random() * 0.1)  # 0-100ms


def _direct_children_with_cmdline(p):
    """Helper to fetch cmdline from children process list"""
    children_with_cmdline = []
    for c in p.children():
        try:
            cmdline = " ".join(c.cmdline())
            if not c.is_running() or not cmdline:
                # Under linux is_running() can return True even though
                # the command line data can no longer be read from
                # /proc/<pid>/cmdline. This looks like a race condition
                # between /proc/<pid>/stat and /proc/<pid>/cmdline
                # when the process is being terminated by the OS.
                continue
            children_with_cmdline.append((c, cmdline))
        except (OSError,) + psutil_exceptions:
            # These errors indicate that the process has terminated while
            # we were processing the info. Just discard it.
            pass
    return children_with_cmdline


def _running_children_pids_with_cmdline(p):
    all_children = _direct_children_with_cmdline(p)
    workers = [
        (c, cmdline)
        for c, cmdline in all_children
        if (
            "semaphore_tracker" not in cmdline
            and "resource_tracker" not in cmdline
            and "multiprocessing.forkserver" not in cmdline
        )
    ]

    forkservers = [
        c
        for c, cmdline in all_children
        if "multiprocessing.forkserver" in cmdline
    ]
    for fs in forkservers:
        workers += _direct_children_with_cmdline(fs)
    return [(w.pid, cmdline) for w, cmdline in workers]


def _check_subprocesses_number(
    executor=None,
    expected_process_number=None,
    expected_max_process_number=None,
    patience=10,
):
    if not psutil:
        # psutil is not installed, we cannot check the number of subprocesses
        return

    for trial_idx in range(patience):
        try:
            # Wait for terminating processes to disappear
            pids_cmdlines = _running_children_pids_with_cmdline(
                psutil.Process()
            )
            children_pids = {pid for pid, _ in pids_cmdlines}
            if executor is not None:
                worker_pids = set(executor._processes.keys())
                # Consistency check: all workers should be in the children list
                assert worker_pids.issubset(children_pids)
            else:
                # Bypass pids checks when executor has been garbage collected
                worker_pids = children_pids
            if expected_process_number is not None:
                assert (
                    len(children_pids) == expected_process_number
                ), pids_cmdlines
                assert (
                    len(worker_pids) == expected_process_number
                ), pids_cmdlines

            if expected_max_process_number is not None:
                assert (
                    len(children_pids) <= expected_max_process_number
                ), pids_cmdlines
                assert (
                    len(worker_pids) <= expected_max_process_number
                ), pids_cmdlines

            return
        except AssertionError as e:
            # Sometimes executor._processes or psutil seems to report
            # out-of-sync information. Let's wait a bit an try again later to see if:
            # - worker_pids is consistent with the psutil children list
            # - the number of children is consistent with the expected (max) number
            if trial_idx == patience - 1:
                raise e
            sleep(0.1)


def _check_executor_started(executor):
    # Submit a small job to make sure that the pool is an working state
    res = executor.submit(id, None)
    try:
        res.result(timeout=TIMEOUT)
    except TimeoutError:
        print(
            "\n" * 3,
            res.done(),
            executor._call_queue.empty(),
            executor._result_queue.empty(),
        )
        print(executor._processes)
        print(threading.enumerate())
        from faulthandler import dump_traceback

        dump_traceback()
        executor.submit(dump_traceback).result(TIMEOUT)
        raise RuntimeError("Executor took too long to run basic task.")


class ExecutorMixin:
    worker_count = 5

    @classmethod
    def setup_class(cls):
        print(f"setup class with {cls.context}")
        global _test_event
        if _test_event is None:
            _test_event = cls.context.Event()

    @classmethod
    def teardown_class(cls):
        print(f"teardown class with {cls.context}")
        global _test_event
        if _test_event is not None:
            _test_event = None

    @pytest.fixture(autouse=True)
    def setup_method(self):
        global _test_event
        assert _test_event is not None
        try:
            self.executor = self.executor_type(
                max_workers=self.worker_count,
                context=self.context,
                initializer=initializer_event,
                initargs=(_test_event,),
            )
        except NotImplementedError as e:
            self.skipTest(str(e))
        _check_executor_started(self.executor)
        _check_subprocesses_number(self.executor, self.worker_count)

    def teardown_method(self, method):
        # Make sure executor is not broken if it should not be
        executor = getattr(self, "executor", None)
        if executor is not None:
            try:
                # old pytest markers:
                expect_broken_pool = hasattr(method, "broken_pool")
                # new pytest markers:
                for mark in getattr(method, "pytestmark", []):
                    if mark.name == "broken_pool":
                        expect_broken_pool = True

                if expect_broken_pool:
                    for _ in range(10):
                        # The executor manager thread can take some time to
                        # mark the executor broken.
                        is_actually_broken = executor._flags.broken is not None
                        if is_actually_broken:
                            break
                        sleep(0.1)
                    else:
                        raise AssertionError(
                            "The executor was not flagged broken at the end of "
                            f" {method.__qualname__} as expected."
                        )
                else:
                    # Check that the executor is not broken right away to avoid
                    # wasting CI time. False negative should be very rare.
                    is_actually_broken = executor._flags.broken is not None
                    assert not is_actually_broken
            finally:
                # Always shutdown the executor, with SIGKILL if the executor
                # is actually broken.
                kill_workers = executor._flags.broken is not None
                t_start = time.time()
                executor.shutdown(wait=True, kill_workers=kill_workers)
                dt = time.time() - t_start
                assert dt < 10, "Executor took too long to shutdown"
                _check_subprocesses_number(executor, 0)

    def _prime_executor(self):
        # Make sure that the executor is ready to do work before running the
        # tests. This should reduce the probability of timeouts in the tests.
        futures = [
            self.executor.submit(time.sleep, 0.1)
            for _ in range(self.worker_count)
        ]
        for f in futures:
            f.result()

    @classmethod
    def check_no_running_workers(cls, patience=5, sleep_duration=0.01):
        if psutil is None:
            return

        deadline = time.time() + patience

        while time.time() <= deadline:
            time.sleep(sleep_duration)
            p = psutil.Process()
            workers = _running_children_pids_with_cmdline(p)
            if not workers:
                return

        # Patience exhausted: log the remaining workers command line and
        # raise error.
        print("Remaining worker processes command lines:", file=sys.stderr)
        for w, cmdline in workers:
            print(w.pid, w.status(), end="\n", file=sys.stderr)
            print(cmdline, end="\n\n", file=sys.stderr)
        raise AssertionError(
            f"Expected no more running worker processes but got {len(workers)}"
            f" after waiting {patience:0.3f}s."
        )


class ReusableExecutorMixin:
    def setup_method(self, method):
        default_start_method = get_context().get_start_method()
        assert default_start_method == "loky", default_start_method
        executor = get_reusable_executor(max_workers=2)
        _check_executor_started(executor)
        # There can be less than 2 workers because of the worker timeout
        _check_subprocesses_number(executor, expected_max_process_number=2)

        # Check that there no other running subprocesses beyond the workers
        # of the reusable executor.
        _check_subprocesses_number(
            executor=None, expected_max_process_number=2
        )

    def teardown_method(self, method):
        """Make sure the executor can be recovered after the tests"""
        executor = get_reusable_executor(max_workers=2)
        assert executor.submit(math.sqrt, 1).result() == 1
        # There can be less than 2 workers because of the worker timeout
        _check_subprocesses_number(executor, expected_max_process_number=2)

        # Check that there no other running subprocesses beyond the workers
        # of the reusable executor.
        _check_subprocesses_number(
            executor=None, expected_max_process_number=2
        )

    @classmethod
    def teardown_class(cls):
        executor = get_reusable_executor(max_workers=2)
        executor.shutdown(wait=True)



###############################################################################
# Launch a subprocess using forkexec and make sure only the needed fd are
# shared in the two process.
#
# author: Thomas Moreau and Olivier Grisel
#
import sys
import os
import subprocess


def fork_exec(cmd, keep_fds, env=None):
    import _posixsubprocess

    # Encoded command args as bytes:
    cmd = [os.fsencode(arg) for arg in cmd]

    # Copy the environment variables to set in the child process (also encoded
    # as bytes).
    env = env or {}
    env = {**os.environ, **env}
    encoded_env = []
    for key, value in env.items():
        encoded_env.append(os.fsencode(f"{key}={value}"))

    # Fds with fileno larger than 3 (stdin=0, stdout=1, stderr=2) are be closed
    # in the child process, except for those passed in keep_fds.
    keep_fds = tuple(sorted(map(int, keep_fds)))
    errpipe_read, errpipe_write = os.pipe()

    if sys.version_info >= (3, 14):
        # Python >= 3.14 removed allow_vfork from _posixsubprocess.fork_exec,
        # see https://github.com/python/cpython/pull/121383
        pgid_to_set = [-1]
        allow_vfork = []
    elif sys.version_info >= (3, 11):
        # Python 3.11 - 3.13 has allow_vfork in _posixsubprocess.fork_exec
        pgid_to_set = [-1]
        allow_vfork = [subprocess._USE_VFORK]
    else:
        # Python < 3.11
        pgid_to_set = []
        allow_vfork = []

    try:
        return _posixsubprocess.fork_exec(
            cmd,  # args
            cmd[0:1],  # executable_list
            True,  # close_fds
            keep_fds,  # pass_fds
            None,  # cwd
            encoded_env,  # env
            -1,  # p2cread
            -1,  # p2cwrite
            -1,  # c2pread
            -1,  # c2pwrite
            -1,  # errread
            -1,  # errwrite
            errpipe_read,  # errpipe_read
            errpipe_write,  # errpipe_write
            False,  # restore_signal
            False,  # call_setsid
            *pgid_to_set,  # pgid_to_set
            None,  # gid
            None,  # extra_groups
            None,  # uid
            -1,  # child_umask
            None,  # preexec_fn
            *allow_vfork,  # extra flag if vfork is available
        )
    finally:
        os.close(errpipe_read)
        os.close(errpipe_write)



import sys
import time
import pytest
import threading
import warnings
import multiprocessing as mp

from loky.backend import get_context
from loky import ProcessPoolExecutor
from loky import get_reusable_executor
from loky.backend.queues import SimpleQueue
from loky.backend.reduction import dumps


class SlowlyPickling:
    """Simulate slowly pickling object, e.g. large numpy array or dict"""

    def __init__(self, delay=1):
        self.delay = delay

    def __reduce__(self):
        time.sleep(self.delay)
        return SlowlyPickling, (self.delay,)


class DelayedSimpleQueue(SimpleQueue):
    def __init__(self, reducers=None, ctx=None, delay=0.1):
        super().__init__(reducers=reducers, ctx=ctx)
        self.out_queue = SimpleQueue(ctx=ctx)
        self._inner_reader = self._reader
        self._reader = self.out_queue._reader

        self._start_thread(delay)

    # Overload _start_thread to correctly call our custom _feed
    def _start_thread(self, delay):
        self._thread = threading.Thread(
            target=DelayedSimpleQueue._feed,
            args=(
                self._rlock,
                self._inner_reader,
                self.out_queue._writer,
                delay,
            ),
            name="QueueDeFeederThread",
        )
        self._thread.daemon = True

        self._thread.start()

    def get(self):
        return self.out_queue.get()

    def close(self):
        self.put(None)
        self._thread.join()

    @staticmethod
    def _feed(readlock, reader, writer, delay):

        PICKLE_NONE = dumps(None)

        while True:
            with readlock:
                res = reader.recv_bytes()
            if res == PICKLE_NONE:
                break
            time.sleep(delay)
            writer.send_bytes(res)
        mp.util.debug("Defeeder clean exit")


class TestTimeoutExecutor:
    def test_worker_timeout_mock(self):
        timeout = 0.001
        context = get_context()
        executor = ProcessPoolExecutor(
            max_workers=4, context=context, timeout=timeout
        )
        result_queue = DelayedSimpleQueue(ctx=context, delay=0.001)
        executor._result_queue = result_queue

        with pytest.warns(
            UserWarning, match="^A worker stopped while some jobs"
        ):
            for _ in range(5):
                # Trigger worker spawn for lazy executor implementations
                for _ in executor.map(id, range(8)):
                    pass

        executor.shutdown()
        result_queue.close()

    def test_worker_timeout_with_slowly_pickling_objects(self, n_tasks=5):
        """Check that the worker timeout can be low without deadlocking

        In particular if dispatching call items to the queue is slow because of
        pickling large arguments, the executor should ensure that there is an
        appropriate amount of workers to move one and not get stalled.
        """
        with pytest.warns(
            UserWarning, match="^A worker stopped while some jobs"
        ):
            for timeout, delay in [
                (0.01, 0.02),
                (0.01, 0.1),
                (0.1, 0.1),
                (0.001, 0.1),
            ]:
                executor = get_reusable_executor(
                    max_workers=2, timeout=timeout
                )
                # First make sure the executor is started
                executor.submit(id, 42).result()
                results = list(
                    executor.map(id, [SlowlyPickling(delay)] * n_tasks)
                )
                assert len(results) == n_tasks

    def test_worker_timeout_shutdown_no_deadlock(self):
        """Check that worker timeout don't cause deadlock when shutting down."""
        with warnings.catch_warnings(record=True) as record:
            with get_reusable_executor(max_workers=2, timeout=0.001) as e:
                # First make sure the executor is started
                e.submit(id, 42).result()

                # Then give a task that will take more time to be pickled so
                # we are sure that the worker timeout.
                f = e.submit(id, SlowlyPickling(1))
        f.result()

        # We can expect to observe warnings related to worker timeout while
        # there are still pending tasks but this should not cause a deadlock.
        for w in record:
            msg = w.message.args[0]
            assert "A worker stopped while some jobs" in msg



from loky import process_executor
from loky.backend import get_context

from ._executor_mixin import ExecutorMixin
from ._test_process_executor import (
    AsCompletedTests,
    ExecutorShutdownTest,
    ExecutorTest,
    WaitTests,
)


class ProcessPoolLokyMixin(ExecutorMixin):
    # Makes sure that the context is defined
    executor_type = process_executor.ProcessPoolExecutor
    context = get_context("loky")


class TestsProcessPoolLokyShutdown(ProcessPoolLokyMixin, ExecutorShutdownTest):
    def _prime_executor(self):
        pass


class TestsProcessPoolLokyWait(ProcessPoolLokyMixin, WaitTests):
    pass


class TestsProcessPoolLokyAsCompleted(ProcessPoolLokyMixin, AsCompletedTests):
    pass


class TestsProcessPoolLokyExecutor(ProcessPoolLokyMixin, ExecutorTest):
    pass



import sys

from loky import process_executor
from loky.backend import get_context

from ._executor_mixin import ExecutorMixin


if sys.platform != "win32":
    from ._test_process_executor import (
        AsCompletedTests,
        ExecutorShutdownTest,
        ExecutorTest,
        WaitTests,
    )

    class ProcessPoolForkserverMixin(ExecutorMixin):
        executor_type = process_executor.ProcessPoolExecutor
        context = get_context("forkserver")

    class TestsProcessPoolForkserverShutdown(
        ProcessPoolForkserverMixin, ExecutorShutdownTest
    ):
        def _prime_executor(self):
            pass

    class TestsProcessPoolForkserverWait(
        ProcessPoolForkserverMixin, WaitTests
    ):
        pass

    class TestsProcessPoolForkserverAsCompleted(
        ProcessPoolForkserverMixin, AsCompletedTests
    ):
        pass

    class TestsProcessPoolForkserverExecutor(
        ProcessPoolForkserverMixin, ExecutorTest
    ):
        pass



###############################################################################
# Basic context management with LokyContext
#
# author: Thomas Moreau and Olivier Grisel
#
# adapted from multiprocessing/context.py
#  * Create a context ensuring loky uses only objects that are compatible
#  * Add LokyContext to the list of context of multiprocessing so loky can be
#    used with multiprocessing.set_start_method
#  * Implement a CFS-aware amd physical-core aware cpu_count function.
#
import os
import sys
import math
import subprocess
import traceback
import warnings
import multiprocessing as mp
from multiprocessing import get_context as mp_get_context
from multiprocessing.context import BaseContext
from concurrent.futures.process import _MAX_WINDOWS_WORKERS


from .process import LokyProcess, LokyInitMainProcess

# Apparently, on older Python versions, loky cannot work 61 workers on Windows
# but instead 60: ¯\_(ツ)_/¯
if sys.version_info < (3, 10):
    _MAX_WINDOWS_WORKERS = _MAX_WINDOWS_WORKERS - 1

START_METHODS = ["loky", "loky_init_main", "spawn"]
if sys.platform != "win32":
    START_METHODS += ["fork", "forkserver"]

_DEFAULT_START_METHOD = None

# Cache for the number of physical cores to avoid repeating subprocess calls.
# It should not change during the lifetime of the program.
physical_cores_cache = None


def get_context(method=None):
    # Try to overload the default context
    method = method or _DEFAULT_START_METHOD or "loky"
    if method == "fork":
        # If 'fork' is explicitly requested, warn user about potential issues.
        warnings.warn(
            "`fork` start method should not be used with "
            "`loky` as it does not respect POSIX. Try using "
            "`spawn` or `loky` instead.",
            UserWarning,
        )
    try:
        return mp_get_context(method)
    except ValueError:
        raise ValueError(
            f"Unknown context '{method}'. Value should be in "
            f"{START_METHODS}."
        )


def set_start_method(method, force=False):
    global _DEFAULT_START_METHOD
    if _DEFAULT_START_METHOD is not None and not force:
        raise RuntimeError("context has already been set")
    assert method is None or method in START_METHODS, (
        f"'{method}' is not a valid start_method. It should be in "
        f"{START_METHODS}"
    )

    _DEFAULT_START_METHOD = method


def get_start_method():
    return _DEFAULT_START_METHOD


def cpu_count(only_physical_cores=False):
    """Return the number of CPUs the current process can use.

    The returned number of CPUs accounts for:
     * the number of CPUs in the system, as given by
       ``multiprocessing.cpu_count``;
     * the CPU affinity settings of the current process
       (available on some Unix systems);
     * Cgroup CPU bandwidth limit (available on Linux only, typically
       set by docker and similar container orchestration systems);
     * the value of the LOKY_MAX_CPU_COUNT environment variable if defined.
    and is given as the minimum of these constraints.

    If ``only_physical_cores`` is True, return the number of physical cores
    instead of the number of logical cores (hyperthreading / SMT). Note that
    this option is not enforced if the number of usable cores is controlled in
    any other way such as: process affinity, Cgroup restricted CPU bandwidth
    or the LOKY_MAX_CPU_COUNT environment variable. If the number of physical
    cores is not found, return the number of logical cores.

    Note that on Windows, the returned number of CPUs cannot exceed 61 (or 60 for
    Python < 3.10), see:
    https://bugs.python.org/issue26903.

    It is also always larger or equal to 1.
    """
    # Note: os.cpu_count() is allowed to return None in its docstring
    os_cpu_count = os.cpu_count() or 1
    if sys.platform == "win32":
        # On Windows, attempting to use more than 61 CPUs would result in a
        # OS-level error. See https://bugs.python.org/issue26903. According to
        # https://learn.microsoft.com/en-us/windows/win32/procthread/processor-groups
        # it might be possible to go beyond with a lot of extra work but this
        # does not look easy.
        os_cpu_count = min(os_cpu_count, _MAX_WINDOWS_WORKERS)

    cpu_count_user = _cpu_count_user(os_cpu_count)
    aggregate_cpu_count = max(min(os_cpu_count, cpu_count_user), 1)

    if not only_physical_cores:
        return aggregate_cpu_count

    if cpu_count_user < os_cpu_count:
        # Respect user setting
        return max(cpu_count_user, 1)

    cpu_count_physical, exception = _count_physical_cores()
    if cpu_count_physical != "not found":
        return cpu_count_physical

    # Fallback to default behavior
    if exception is not None:
        # warns only the first time
        warnings.warn(
            "Could not find the number of physical cores for the "
            f"following reason:\n{exception}\n"
            "Returning the number of logical cores instead. You can "
            "silence this warning by setting LOKY_MAX_CPU_COUNT to "
            "the number of cores you want to use."
        )
        traceback.print_tb(exception.__traceback__)

    return aggregate_cpu_count


def _cpu_count_cgroup(os_cpu_count):
    # Cgroup CPU bandwidth limit available in Linux since 2.6 kernel
    cpu_max_fname = "/sys/fs/cgroup/cpu.max"
    cfs_quota_fname = "/sys/fs/cgroup/cpu/cpu.cfs_quota_us"
    cfs_period_fname = "/sys/fs/cgroup/cpu/cpu.cfs_period_us"
    if os.path.exists(cpu_max_fname):
        # cgroup v2
        # https://www.kernel.org/doc/html/latest/admin-guide/cgroup-v2.html
        with open(cpu_max_fname) as fh:
            cpu_quota_us, cpu_period_us = fh.read().strip().split()
    elif os.path.exists(cfs_quota_fname) and os.path.exists(cfs_period_fname):
        # cgroup v1
        # https://www.kernel.org/doc/html/latest/scheduler/sched-bwc.html#management
        with open(cfs_quota_fname) as fh:
            cpu_quota_us = fh.read().strip()
        with open(cfs_period_fname) as fh:
            cpu_period_us = fh.read().strip()
    else:
        # No Cgroup CPU bandwidth limit (e.g. non-Linux platform)
        cpu_quota_us = "max"
        cpu_period_us = 100_000  # unused, for consistency with default values

    if cpu_quota_us == "max":
        # No active Cgroup quota on a Cgroup-capable platform
        return os_cpu_count
    else:
        cpu_quota_us = int(cpu_quota_us)
        cpu_period_us = int(cpu_period_us)
        if cpu_quota_us > 0 and cpu_period_us > 0:
            return math.ceil(cpu_quota_us / cpu_period_us)
        else:  # pragma: no cover
            # Setting a negative cpu_quota_us value is a valid way to disable
            # cgroup CPU bandwith limits
            return os_cpu_count


def _cpu_count_affinity(os_cpu_count):
    # Number of available CPUs given affinity settings
    if hasattr(os, "sched_getaffinity"):
        try:
            return len(os.sched_getaffinity(0))
        except NotImplementedError:
            pass

    # On some platforms, os.sched_getaffinity does not exist or raises
    # NotImplementedError, let's try with the psutil if installed.
    try:
        import psutil

        p = psutil.Process()
        if hasattr(p, "cpu_affinity"):
            return len(p.cpu_affinity())

    except ImportError:  # pragma: no cover
        if (
            sys.platform == "linux"
            and os.environ.get("LOKY_MAX_CPU_COUNT") is None
        ):
            # Some platforms don't implement os.sched_getaffinity on Linux which
            # can cause severe oversubscription problems. Better warn the
            # user in this particularly pathological case which can wreck
            # havoc, typically on CI workers.
            warnings.warn(
                "Failed to inspect CPU affinity constraints on this system. "
                "Please install psutil or explictly set LOKY_MAX_CPU_COUNT."
            )

    # This can happen for platforms that do not implement any kind of CPU
    # infinity such as macOS-based platforms.
    return os_cpu_count


def _cpu_count_user(os_cpu_count):
    """Number of user defined available CPUs"""
    cpu_count_affinity = _cpu_count_affinity(os_cpu_count)

    cpu_count_cgroup = _cpu_count_cgroup(os_cpu_count)

    # User defined soft-limit passed as a loky specific environment variable.
    cpu_count_loky = int(os.environ.get("LOKY_MAX_CPU_COUNT", os_cpu_count))

    return min(cpu_count_affinity, cpu_count_cgroup, cpu_count_loky)


def _count_physical_cores():
    """Return a tuple (number of physical cores, exception)

    If the number of physical cores is found, exception is set to None.
    If it has not been found, return ("not found", exception).

    The number of physical cores is cached to avoid repeating subprocess calls.
    """
    exception = None

    # First check if the value is cached
    global physical_cores_cache
    if physical_cores_cache is not None:
        return physical_cores_cache, exception

    # Not cached yet, find it
    try:
        if sys.platform == "linux":
            cpu_count_physical = _count_physical_cores_linux()
        elif sys.platform == "win32":
            cpu_count_physical = _count_physical_cores_win32()
        elif sys.platform == "darwin":
            cpu_count_physical = _count_physical_cores_darwin()
        else:
            raise NotImplementedError(f"unsupported platform: {sys.platform}")

        # if cpu_count_physical < 1, we did not find a valid value
        if cpu_count_physical < 1:
            raise ValueError(f"found {cpu_count_physical} physical cores < 1")

    except Exception as e:
        exception = e
        cpu_count_physical = "not found"

    # Put the result in cache
    physical_cores_cache = cpu_count_physical

    return cpu_count_physical, exception


def _count_physical_cores_linux():
    try:
        cpu_info = subprocess.run(
            "lscpu --parse=core".split(), capture_output=True, text=True
        )
        cpu_info = cpu_info.stdout.splitlines()
        cpu_info = {line for line in cpu_info if not line.startswith("#")}
        return len(cpu_info)
    except:
        pass  # fallback to /proc/cpuinfo

    cpu_info = subprocess.run(
        "cat /proc/cpuinfo".split(), capture_output=True, text=True
    )
    cpu_info = cpu_info.stdout.splitlines()
    cpu_info = {line for line in cpu_info if line.startswith("core id")}
    return len(cpu_info)


def _count_physical_cores_win32():
    try:
        cmd = "-Command (Get-CimInstance -ClassName Win32_Processor).NumberOfCores"
        cpu_info = subprocess.run(
            f"powershell.exe {cmd}".split(),
            capture_output=True,
            text=True,
        )
        cpu_info = cpu_info.stdout.splitlines()
        return int(cpu_info[0])
    except:
        pass  # fallback to wmic (older Windows versions; deprecated now)

    cpu_info = subprocess.run(
        "wmic CPU Get NumberOfCores /Format:csv".split(),
        capture_output=True,
        text=True,
    )
    cpu_info = cpu_info.stdout.splitlines()
    cpu_info = [
        l.split(",")[1] for l in cpu_info if (l and l != "Node,NumberOfCores")
    ]
    return sum(map(int, cpu_info))


def _count_physical_cores_darwin():
    cpu_info = subprocess.run(
        "sysctl -n hw.physicalcpu".split(),
        capture_output=True,
        text=True,
    )
    cpu_info = cpu_info.stdout
    return int(cpu_info)


class LokyContext(BaseContext):
    """Context relying on the LokyProcess."""

    _name = "loky"
    Process = LokyProcess
    cpu_count = staticmethod(cpu_count)

    def Queue(self, maxsize=0, reducers=None):
        """Returns a queue object"""
        from .queues import Queue

        return Queue(maxsize, reducers=reducers, ctx=self.get_context())

    def SimpleQueue(self, reducers=None):
        """Returns a queue object"""
        from .queues import SimpleQueue

        return SimpleQueue(reducers=reducers, ctx=self.get_context())

    if sys.platform != "win32":
        """For Unix platform, use our custom implementation of synchronize
        ensuring that we use the loky.backend.resource_tracker to clean-up
        the semaphores in case of a worker crash.
        """

        def Semaphore(self, value=1):
            """Returns a semaphore object"""
            from .synchronize import Semaphore

            return Semaphore(value=value)

        def BoundedSemaphore(self, value):
            """Returns a bounded semaphore object"""
            from .synchronize import BoundedSemaphore

            return BoundedSemaphore(value)

        def Lock(self):
            """Returns a lock object"""
            from .synchronize import Lock

            return Lock()

        def RLock(self):
            """Returns a recurrent lock object"""
            from .synchronize import RLock

            return RLock()

        def Condition(self, lock=None):
            """Returns a condition object"""
            from .synchronize import Condition

            return Condition(lock)

        def Event(self):
            """Returns an event object"""
            from .synchronize import Event

            return Event()


class LokyInitMainContext(LokyContext):
    """Extra context with LokyProcess, which does load the main module

    This context is used for compatibility in the case ``cloudpickle`` is not
    present on the running system. This permits to load functions defined in
    the ``main`` module, using proper safeguards. The declaration of the
    ``executor`` should be protected by ``if __name__ == "__main__":`` and the
    functions and variable used from main should be out of this block.

    This mimics the default behavior of multiprocessing under Windows and the
    behavior of the ``spawn`` start method on a posix system.
    For more details, see the end of the following section of python doc
    https://docs.python.org/3/library/multiprocessing.html#multiprocessing-programming
    """

    _name = "loky_init_main"
    Process = LokyInitMainProcess


# Register loky context so it works with multiprocessing.get_context
ctx_loky = LokyContext()
mp.context._concrete_contexts["loky"] = ctx_loky
mp.context._concrete_contexts["loky_init_main"] = LokyInitMainContext()



import multiprocessing as mp
import os
import sys
import shutil
import subprocess
import tempfile
import warnings
from subprocess import check_output

import pytest

import loky
from loky import cpu_count
from loky.backend.context import _cpu_count_user, _MAX_WINDOWS_WORKERS


def test_version():
    assert hasattr(
        loky, "__version__"
    ), "There are no __version__ argument on the loky module"


def test_cpu_count(monkeypatch):

    # Monkeypatch subprocess.run to simulate the absence of lscpu on linux or CIM on
    # windows to test the different code paths in _cpu_count_physical.
    old_run = subprocess.run

    def mock_run(*args, **kwargs):
        if (
            "lscpu" in args[0]
            and os.environ.get("LOKY_TEST_NO_LSCPU") == "true"
        ):
            raise RuntimeError("lscpu not available")

        if (
            "powershell.exe" in args[0]
            and os.environ.get("LOKY_TEST_NO_CIM") == "true"
        ):
            raise RuntimeError("Cim not available")

        return old_run(*args, **kwargs)

    monkeypatch.setattr(subprocess, "run", mock_run)

    cpus = cpu_count()
    assert type(cpus) is int
    assert cpus >= 1

    cpus_physical = cpu_count(only_physical_cores=True)
    assert type(cpus_physical) is int
    assert 1 <= cpus_physical <= cpus

    # again to check that it's correctly cached
    cpus_physical = cpu_count(only_physical_cores=True)
    assert type(cpus_physical) is int
    assert 1 <= cpus_physical <= cpus


@pytest.mark.skipif(sys.platform != "win32", reason="Windows specific test")
def test_windows_max_cpu_count():
    assert cpu_count() <= _MAX_WINDOWS_WORKERS


cpu_count_cmd = (
    "from loky.backend.context import cpu_count;" "print(cpu_count({args}))"
)


def test_cpu_count_os_sched_getaffinity():
    if not hasattr(os, "sched_getaffinity") or not hasattr(shutil, "which"):
        pytest.skip()

    taskset_bin = shutil.which("taskset")
    python_bin = shutil.which("python")

    if taskset_bin is None or python_bin is None:
        raise pytest.skip()

    try:
        os.sched_getaffinity(0)
    except NotImplementedError:
        pytest.skip()

    res = check_output(
        [
            taskset_bin,
            "-c",
            "0",
            python_bin,
            "-c",
            cpu_count_cmd.format(args=""),
        ],
        text=True,
    )

    res_physical = check_output(
        [
            taskset_bin,
            "-c",
            "0",
            python_bin,
            "-c",
            cpu_count_cmd.format(args="only_physical_cores=True"),
        ],
        text=True,
    )

    assert res.strip() == "1"
    assert res_physical.strip() == "1"


def test_cpu_count_psutil_affinity():
    psutil = pytest.importorskip("psutil")
    p = psutil.Process()
    if not hasattr(p, "cpu_affinity"):
        pytest.skip("psutil does not provide cpu_affinity on this platform")

    original_affinity = p.cpu_affinity()
    assert cpu_count() <= len(original_affinity)
    try:
        new_affinity = original_affinity[:1]
        p.cpu_affinity(new_affinity)
        assert cpu_count() == 1
    finally:
        p.cpu_affinity(original_affinity)


def test_cpu_count_cgroup_limit():
    if sys.platform == "win32":
        pytest.skip()

    if not hasattr(shutil, "which"):
        pytest.skip()

    docker_bin = shutil.which("docker")
    if docker_bin is None:
        raise pytest.skip("docker is required to run this test")

    loky_module_path = os.path.abspath(os.path.dirname(loky.__file__))
    loky_project_path = os.path.abspath(
        os.path.join(loky_module_path, os.pardir)
    )

    # The following will always run using the Python 3.7 docker image.
    # We mount the loky source as /loky inside the container,
    # so it can be imported when running commands under /

    # Tell docker to configure the Cgroup quota to use 0.5 CPU, loky will
    # always detect 1 CPU because it rounds up to the next integer.
    res_500_mCPU = int(
        check_output(
            f"{docker_bin} run --rm --cpus 0.5 -v {loky_project_path}:/loky python:3.10 "
            f"/bin/bash -c 'pip install --quiet -e /loky ; "
            f"python -c \"{cpu_count_cmd.format(args='')}\"'",
            shell=True,
            text=True,
        ).strip()
    )
    assert res_500_mCPU == 1

    # Limiting to 1.5 CPUs can lead to 1 if there is only 1 CPU on the machine or
    # 2 if there are 2 CPUs or more.
    res_1500_mCPU = int(
        check_output(
            f"{docker_bin} run --rm --cpus 1.5 -v {loky_project_path}:/loky python:3.10 "
            f"/bin/bash -c 'pip install --quiet -e /loky ; "
            f"python -c \"{cpu_count_cmd.format(args='')}\"'",
            shell=True,
            text=True,
        ).strip()
    )
    assert res_1500_mCPU in (1, 2)

    # By default there is no limit: use all available CPUs.
    res_default = int(
        check_output(
            f"{docker_bin} run --rm -v {loky_project_path}:/loky python:3.10 "
            f"/bin/bash -c 'pip install --quiet -e /loky ; "
            f"python -c \"{cpu_count_cmd.format(args='')}\"'",
            shell=True,
            text=True,
        ).strip()
    )
    assert res_default >= res_1500_mCPU


def test_only_physical_cores_error():
    # Check the warning issued by cpu_count(only_physical_cores=True) when
    # unable to retrieve the number of physical cores.
    if sys.platform != "linux":
        pytest.skip()

    # if number of available cpus is already restricted, cpu_count will return
    # that value and no warning is issued even if only_physical_cores == True.
    # (tested in another test: test_only_physical_cores_with_user_limitation
    cpu_count_mp = mp.cpu_count()
    if _cpu_count_user(cpu_count_mp) < cpu_count_mp:
        pytest.skip()

    with tempfile.TemporaryDirectory() as tmp_dir:
        # Write bad lscpu program
        lscpu_path = f"{tmp_dir}/lscpu"
        with open(lscpu_path, "w") as f:
            f.write("#!/bin/sh\n" "exit(1)")
        os.chmod(lscpu_path, 0o777)

        try:
            old_path = os.environ["PATH"]
            os.environ["PATH"] = f"{tmp_dir}:{old_path}"

            # clear the cache otherwise the warning is not triggered
            import loky.backend.context

            loky.backend.context.physical_cores_cache = None

            with pytest.warns(
                UserWarning,
                match="Could not find the number of" " physical cores",
            ):
                cpu_count(only_physical_cores=True)

            # Should not warn the second time
            with warnings.catch_warnings():
                warnings.simplefilter("error")
                cpu_count(only_physical_cores=True)

        finally:
            os.environ["PATH"] = old_path


def test_only_physical_cores_with_user_limitation():
    # Check that user limitation for the available number of cores is
    # respected even if only_physical_cores == True
    cpu_count_mp = mp.cpu_count()
    cpu_count_user = _cpu_count_user(cpu_count_mp)

    if cpu_count_user < cpu_count_mp:
        assert cpu_count() == cpu_count_user
        assert cpu_count(only_physical_cores=True) == cpu_count_user
